# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import logging
import re
import urllib.parse
from contextlib import suppress
from pathlib import Path

from gitfourchette import settings
from gitfourchette.forms.brandeddialog import convertToBrandedDialog
from gitfourchette.forms.ui_clonedialog import Ui_CloneDialog
from gitfourchette.gitdriver import argsIf
from gitfourchette.localization import *
from gitfourchette.porcelain import RepoContext, RepositoryOpenFlag
from gitfourchette.qt import *
from gitfourchette.repoprefs import RepoPrefs
from gitfourchette.tasks import RepoTask, RepoTaskRunner
from gitfourchette.toolbox import *

logger = logging.getLogger(__name__)


def _projectNameFromUrl(url: str) -> str:
    url = url.strip()
    name = url.rsplit("/", 1)[-1].removesuffix(".git")
    name = urllib.parse.unquote(name)
    # Sanitize name
    for c in " ?/\\*~<>|:":
        name = name.replace(c, "_")
    return name


class CloneDialog(QDialog):
    cloneSuccessful = Signal(str)

    urlEditUserDataClearHistory = "CLEAR_HISTORY"

    def __init__(self, initialUrl: str, parent: QWidget):
        super().__init__(parent)

        if not initialUrl:
            initialUrl = guessRemoteUrlFromText(QApplication.clipboard().text())

        self.taskRunner = RepoTaskRunner(self)

        self.ui = Ui_CloneDialog()
        self.ui.setupUi(self)

        self.initUrlComboBox()
        self.ui.urlEdit.activated.connect(self.onUrlActivated)

        self.ui.browseButton.setIcon(stockIcon("SP_DialogOpenButton"))
        self.ui.browseButton.clicked.connect(self.browse)

        self.setDefaultCloneLocationAction = QAction("(SET)")
        self.setDefaultCloneLocationAction.triggered.connect(lambda: self.setDefaultCloneLocationPref(self.pathParentDir))
        self.clearDefaultCloneLocationAction = QAction(stockIcon("edit-clear-history"), _("Reset default clone location"))
        self.clearDefaultCloneLocationAction.triggered.connect(lambda: self.setDefaultCloneLocationPref(""))
        self.ui.browseButton.setMenu(ActionDef.makeQMenu(
            self.ui.browseButton, [self.setDefaultCloneLocationAction, self.clearDefaultCloneLocationAction]))
        self.updateDefaultCloneLocationAction()  # prime default clone path actions
        self.ui.pathEdit.textChanged.connect(self.updateDefaultCloneLocationAction)

        self.cloneButton: QPushButton = self.ui.buttonBox.button(QDialogButtonBox.StandardButton.Ok)
        self.cloneButton.setText(_("C&lone"))
        self.cloneButton.setIcon(QIcon.fromTheme("download"))
        self.cloneButton.clicked.connect(self.onCloneClicked)

        self.cancelButton: QPushButton = self.ui.buttonBox.button(QDialogButtonBox.StandardButton.Cancel)
        self.cancelButton.setAutoDefault(False)

        self.ui.statusForm.connectAbortButton(self.cancelButton)
        self.ui.statusForm.setBlurb(_("Hit {0} when ready.", tquo(self.cloneButton.text().replace("&", ""))))

        self.ui.shallowCloneDepthSpinBox.valueChanged.connect(self.onShallowCloneDepthChanged)
        self.ui.shallowCloneCheckBox.checkStateChanged.connect(self.onShallowCloneCheckStateChanged)
        self.ui.shallowCloneCheckBox.setMinimumHeight(max(self.ui.shallowCloneCheckBox.height(), self.ui.shallowCloneDepthSpinBox.height()))  # prevent jumping around
        self.onShallowCloneCheckStateChanged(self.ui.shallowCloneCheckBox.checkState())

        convertToBrandedDialog(self)

        self.ui.urlEdit.editTextChanged.connect(self.autoFillDownloadPath)
        self.ui.urlEdit.setCurrentIndex(-1)  # prevent "Clear History" from installing its icon
        self.ui.urlEdit.setCurrentText(initialUrl)
        self.ui.urlEdit.setFocus()

        # Connect protocol button to URL editor
        self.ui.protocolButton.connectTo(self.ui.urlEdit.lineEdit())

        validator = ValidatorMultiplexer(self)
        validator.setGatedWidgets(self.cloneButton)
        validator.connectInput(self.ui.urlEdit.lineEdit(), self.validateUrl)
        validator.connectInput(self.ui.pathEdit, self.validatePath)
        validator.run(silenceEmptyWarnings=True)

    def validateUrl(self, url):
        if not url:
            return _("Please fill in this field.")
        return ""

    def validatePath(self, _ignored: str) -> str:
        path = self.path
        path = Path(path)
        if not path.is_absolute():
            return _("Please enter an absolute path.")
        if path.is_file():
            return _("There’s already a file at this path.")
        if path.is_dir():
            with suppress(StopIteration):
                next(path.iterdir())  # raises StopIteration if directory is not empty
                return _("This directory isn’t empty.")
        return ""

    def autoFillDownloadPath(self, url):
        # Get standard download location
        downloadPath = Path(settings.prefs.resolveDefaultCloneLocation())

        # Don't overwrite if user has set a custom path
        currentPath = self.ui.pathEdit.text()
        if currentPath and Path(currentPath).parent != downloadPath:
            return

        # Extract project name; clear target path if blank
        projectName = _projectNameFromUrl(url)
        if not projectName or projectName in [".", ".."]:
            self.ui.pathEdit.setText("")
            return

        # Append differentiating number if this path already exists
        projectName = withUniqueSuffix(projectName, lambda x: (downloadPath / x).exists())

        # Set target path to <downloadPath>/<projectName>
        target = downloadPath / projectName
        assert not target.exists()

        self.ui.pathEdit.setText(str(target))

    def updateDefaultCloneLocationAction(self):
        self.clearDefaultCloneLocationAction.setEnabled(settings.prefs.defaultCloneLocation != "")

        location = self.pathParentDir
        action = self.setDefaultCloneLocationAction

        if self.validatePath(self.path):  # truthy if validation error
            action.setText(_("Set current location as default clone location"))
            action.setEnabled(False)
            return

        display = lquoe(compactPath(location))
        if location == settings.prefs.resolveDefaultCloneLocation():
            action.setEnabled(False)
            action.setText(_("{0} is the default clone location", display))
        else:
            action.setEnabled(True)
            action.setText(_("Set {0} as default clone location", display))

    def setDefaultCloneLocationPref(self, location: str):
        settings.prefs.defaultCloneLocation = location
        settings.prefs.setDirty()
        self.updateDefaultCloneLocationAction()

    def initUrlComboBox(self):
        urlEdit = self.ui.urlEdit
        urlEdit.clear()

        if settings.history.cloneHistory:
            for url in settings.history.cloneHistory:
                urlEdit.addItem(url)
            urlEdit.insertSeparator(urlEdit.count())

        urlEdit.addItem(stockIcon("edit-clear-history"), _("Clear history"), CloneDialog.urlEditUserDataClearHistory)

        # "Clear history" is added even if the history is empty, so that the
        # QComboBox's arrow button - which cannot be hidden - still pops up
        # something when clicked, as the user might expect.
        if not settings.history.cloneHistory:
            itemModel = urlEdit.model()
            assert isinstance(itemModel, QStandardItemModel)
            clearItem: QStandardItem = itemModel.item(urlEdit.count() - 1)
            clearItem.setFlags(clearItem.flags() & ~Qt.ItemFlag.ItemIsEnabled)

        self.ui.urlEdit.setSizeAdjustPolicy(QComboBox.SizeAdjustPolicy.AdjustToMinimumContentsLengthWithIcon)

    def onUrlActivated(self, index: int):
        itemData = self.ui.urlEdit.itemData(index, Qt.ItemDataRole.UserRole)
        if itemData == CloneDialog.urlEditUserDataClearHistory:
            settings.history.clearCloneHistory()
            settings.history.write()
            self.initUrlComboBox()

    @DisableWidgetUpdatesContext.methodDecorator
    def onShallowCloneCheckStateChanged(self, state: Qt.CheckState):
        isChecked = state == Qt.CheckState.Checked
        if isChecked:
            self.onShallowCloneDepthChanged(self.ui.shallowCloneDepthSpinBox.value())
        else:
            self.ui.shallowCloneCheckBox.setText(_("&Shallow clone"))
        self.ui.shallowCloneDepthSpinBox.setVisible(isChecked)
        self.ui.shallowCloneSuffix.setVisible(isChecked)

    def onShallowCloneDepthChanged(self, depth: int):
        # Re-translate text for correct plural form
        text = _n("&Shallow clone: Fetch up to {n} commit per branch",
                  "&Shallow clone: Fetch up to {n} commits per branch", depth)
        parts = re.split(r"\d(?:.*\d)?", text, maxsplit=1)
        assert len(parts) >= 2
        self.ui.shallowCloneCheckBox.setText(parts[0].strip())
        self.ui.shallowCloneSuffix.setText(parts[1].strip())

    def reject(self):  # connected to abort button and ESC key
        if self.taskRunner.isBusy():
            self.ui.statusForm.requestAbort()
        else:
            super().reject()  # destroys the dialog then emits the "rejected" signal

    @property
    def url(self):
        return self.ui.urlEdit.currentText().strip()

    @property
    def path(self):
        text = self.ui.pathEdit.text().strip()
        path = Path(text)
        with suppress(RuntimeError):
            path = path.expanduser()
        return str(path)

    @property
    def pathParentDir(self):
        return str(Path(self.path).parent)

    def browse(self):
        existingPath = Path(self.path) if self.path else None

        if existingPath:
            initialName = existingPath.name
        else:
            initialName = _projectNameFromUrl(self.url)

        qfd = PersistentFileDialog.saveFile(self, "NewRepo", _("Clone repository into"), initialName)

        # Rationale for omitting directory-related options that appear to make sense at first glance:
        # - FileMode.Directory: forces user to hit "new folder" to enter the name of the repo
        # - Options.ShowDirsOnly: KDE Plasma 6's native dialog forces user to hit "new folder" when this flag is set
        qfd.setOption(QFileDialog.Option.DontConfirmOverwrite, True)  # we'll show our own warning if the file already exists
        qfd.setOption(QFileDialog.Option.HideNameFilterDetails, True)  # not sure Qt honors this...

        qfd.setLabelText(QFileDialog.DialogLabel.FileName, self.ui.pathLabel.text())  # "Clone into:"
        qfd.setLabelText(QFileDialog.DialogLabel.Accept, _("Clone here"))

        if existingPath:
            qfd.setDirectory(str(existingPath.parent))

        qfd.fileSelected.connect(lambda path: self.ui.pathEdit.setText(str(Path(path))))
        qfd.show()

    def enableInputs(self, enable):
        grayable = [
            self.ui.urlLabel,
            self.ui.urlEdit,
            self.ui.pathLabel,
            self.ui.pathEdit,
            self.ui.browseButton,
            self.ui.optionsLabel,
            self.ui.recurseSubmodulesCheckBox,
            self.ui.shallowCloneCheckBox,
            self.ui.shallowCloneDepthSpinBox,
            self.ui.shallowCloneSuffix,
            self.ui.keyFilePicker,
            self.ui.protocolButton,
            self.cloneButton
        ]
        for widget in grayable:
            widget.setEnabled(enable)

    def onCloneClicked(self):
        depth = 0
        privKeyPath = self.ui.keyFilePicker.privateKeyPath()
        recursive = self.ui.recurseSubmodulesCheckBox.isChecked()

        if self.ui.shallowCloneCheckBox.isChecked():
            depth = self.ui.shallowCloneDepthSpinBox.value()

        self.ui.statusForm.initProgress(_("Contacting remote host…"))
        CloneTask.invoke(self,
                         dialog=self, url=self.url, path=self.path, depth=depth,
                         privKeyPath=privKeyPath, recursive=recursive)

    def onUrlProtocolChanged(self, newUrl: str):
        # This pushes the new text to the QLineEdit's undo stack (whereas setText clears the undo stack).
        self.ui.urlEdit.lineEdit().selectAll()
        self.ui.urlEdit.lineEdit().insert(newUrl)

    def onCloneSuccessful(self):
        self.cloneSuccessful.emit(self.path)
        self.accept()


class CloneTask(RepoTask):
    """
    Even though we don't have a Repository yet, this is a RepoTask so we can
    easily run the clone operation in a background thread.
    """

    def flow(self, dialog: CloneDialog, url: str, path: str, depth: int, privKeyPath: str, recursive: bool):
        shallow = depth != 0

        dialog.enableInputs(False)

        # Clone the repo
        driver = self.createGitProcess(
            "clone",
            "--progress",
            *argsIf(recursive, "--recurse-submodules"),
            *argsIf(shallow and recursive, "--shallow-submodules"),
            *argsIf(shallow, "--depth", str(depth)),
            *argsIf(shallow, "--no-single-branch", "--no-tags"),  # match what libgit2 backend does
            "--",
            url,
            path,
            customKey=privKeyPath)
        dialog.ui.statusForm.connectProcess(driver)
        yield from self.flowStartProcess(driver, autoFail=False)

        if driver.exitCode() != 0:
            QApplication.beep()
            QApplication.alert(dialog, 500)
            dialog.enableInputs(True)
            dialog.ui.statusForm.setBlurb(driver.htmlErrorText())
            return

        # Successfully cloned
        settings.history.addCloneUrl(url)
        settings.history.setDirty()

        # Store custom key (if any) in cloned repo config
        if privKeyPath:
            with RepoContext(path, RepositoryOpenFlag.NO_SEARCH) as repo:
                repoPrefs = RepoPrefs.initForRepo(repo)
                repoPrefs.customKeyFile = privKeyPath
                repoPrefs.setDirty()
                repoPrefs.write()

        # When the task runner wraps up, tell dialog to finish
        dialog.taskRunner.ready.connect(dialog.onCloneSuccessful)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from collections.abc import Callable

from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.toolbox import *

HORIZONTAL_CONTENT_HEIGHT = 20
FONT_POINT_PERCENT = 90
PERMANENT_PROPERTY = "permanent"


class Banner(QFrame):
    buttons: list[QToolButton]

    def __init__(self, parent, orientation: Qt.Orientation):
        super().__init__(parent)
        self.setObjectName("Banner")

        icon = QLabel(self)
        icon.setPixmap(stockIcon("SP_MessageBoxInformation").pixmap(min(16, HORIZONTAL_CONTENT_HEIGHT)))

        label = QLabel(__name__, self)
        label.setWordWrap(True)

        layout: QBoxLayout
        if orientation == Qt.Orientation.Vertical:
            layout = QVBoxLayout(self)
        else:
            label.setSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Minimum)
            layout = QHBoxLayout(self)

        layout.setContentsMargins(QMargins())
        layout.addWidget(icon)
        layout.addWidget(label)

        self.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)

        self.icon = icon
        self.label = label
        self.lastWarningWasDismissed = False
        self.orientation = orientation
        self.buttons = []

        self.dismissButton = self.addButton(_("Dismiss"), self.dismiss, permanent=True)

    def addButton(self, text: str, callback: Callable | None = None, permanent=False) -> QToolButton:
        button = QToolButton(self)
        button.setText(text)
        button.setProperty(PERMANENT_PROPERTY, "true" if permanent else "")

        # Harmonize button width in vertical banner
        if self.orientation == Qt.Orientation.Vertical:
            button.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred)

        if self.property("class") != "merge":
            button.setAutoRaise(True)

        self.buttons.append(button)

        if callback is not None:
            button.clicked.connect(callback)

        if self.orientation == Qt.Orientation.Horizontal:
            tweakWidgetFont(button, FONT_POINT_PERCENT)
            button.setMaximumHeight(HORIZONTAL_CONTENT_HEIGHT)

        layout = self.layout()
        assert isinstance(layout, QBoxLayout)
        layout.insertWidget(2, button)
        return button

    def clearButtons(self):
        for i in range(len(self.buttons) - 1, -1, -1):
            button = self.buttons[i]
            if not button.property(PERMANENT_PROPERTY):
                button.hide()
                button.deleteLater()
                del self.buttons[i]

    def popUp(self, title: str, text: str, heeded=False, canDismiss=False, withIcon=False):
        self.clearButtons()
        toggleQssProperty(self, "heeded", heeded)

        smallPt = adjustedWidgetFontSize(self.label, FONT_POINT_PERCENT)
        markup = f"<style>sm {{ font-size: {smallPt}pt; }}</style>"
        if title:
            markup += f"<b>{title}</b>"
            if text:
                markup += f"<br><sm>{text}</sm>"
        else:
            markup += f"<sm>{text}</sm>"

        self.label.setTextFormat(Qt.TextFormat.RichText)
        self.label.setText(markup)
        self.icon.setVisible(withIcon)

        self.dismissButton.setVisible(canDismiss)
        self.lastWarningWasDismissed = False

        self.setVisible(True)

    def dismiss(self):
        self.lastWarningWasDismissed = True
        self.hide()
        self.clearButtons()



# -----------------------------------------------------------------------------
# Copyright (C) 2025 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.graph.graph import (
    Arc,
    ArcJunction,
    BatchRow,
    ChainHandle,
    Frame,
    Graph,
    KF_INTERVAL,
    PlaybackState,
)
from gitfourchette.graph.graphtrickle import GraphTrickle
from gitfourchette.graph.graphdiagram import GraphDiagram
from gitfourchette.graph.graphbuilder import (
    GraphBuildLoop,
    GraphSpliceLoop,
    MockCommit,
    MockOid,
)
from gitfourchette.graph.graphweaver import GraphWeaver



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.forms.ui_checkoutcommitdialog import Ui_CheckoutCommitDialog
from gitfourchette.localization import *
from gitfourchette.porcelain import Oid, RefPrefix
from gitfourchette.qt import *
from gitfourchette.toolbox import *


class CheckoutCommitDialog(QDialog):
    def __init__(
            self,
            oid: Oid,
            refs: list[str],
            currentBranch: str,
            anySubmodules: bool,
            parent=None):

        super().__init__(parent)

        localBranches = [r.removeprefix(RefPrefix.HEADS) for r in refs if r.startswith(RefPrefix.HEADS)]
        currentBranchIsHere = currentBranch in localBranches
        currentBranchIsOnlyLocalBranch = currentBranchIsHere and len(localBranches) == 1
        currentBranchIsOnlyRef = currentBranchIsHere and len(refs) == 1
        isDetachedHead = currentBranch == "HEAD"

        ui = Ui_CheckoutCommitDialog()
        self.ui = ui
        self.ui.setupUi(self)

        self.setWindowTitle(_("Check out commit {0}", shortHash(oid)))

        if isDetachedHead:
            ui.detachHeadRadioButton.setText(_("Move &detached HEAD here"))
        else:
            ui.resetHeadRadioButton.setText(_("&Reset the tip of branch {0} here…", lquo(currentBranch)))
            ui.mergeRadioButton.setText(_("&Merge into {0}…", lquo(currentBranch)))

        self.bindRadioToOkCaption(ui.detachHeadRadioButton, _("Detach HEAD"), "git-head-detached")
        self.bindRadioToOkCaption(ui.switchRadioButton, _("Switch Branch"), "git-checkout")
        self.bindRadioToOkCaption(ui.resetHeadRadioButton, _("Reset {0}…", "HEAD" if isDetachedHead else lquoe(currentBranch)), "")
        self.bindRadioToOkCaption(ui.mergeRadioButton, _("Merge…"), "git-merge")
        self.bindRadioToOkCaption(ui.createBranchRadioButton, _("Create Branch…"), "git-branch")

        # Determine if we can merge
        ui.mergeRadioButton.setEnabled(bool(refs) and not currentBranchIsOnlyRef)

        # Determine if we can switch to a local branch
        switchSuffix = ""
        if localBranches and not currentBranchIsOnlyLocalBranch:
            ui.switchRadioButton.click()

            branchChoices = [r for r in localBranches if r != currentBranch]
            ui.switchComboBox.addItems(branchChoices)

            # Single branch - remove combobox
            if len(branchChoices) == 1:
                switchSuffix = lquo(branchChoices[0])
        else:
            ui.detachHeadRadioButton.click()
            ui.switchRadioButton.setEnabled(False)

            if currentBranchIsOnlyLocalBranch:
                switchSuffix = _p("checkout: current branch name", "{0} (already checked out)", lquo(currentBranch))
            else:
                switchSuffix = _p("checkout: no branches available", "(none available here)")

        # If there's a suffix text for the switch radio button, hide the combobox
        if switchSuffix:
            ui.switchRadioButton.setText(ui.switchRadioButton.text() + " " + switchSuffix)
            ui.switchComboBox.setVisible(False)

        if not anySubmodules:
            ui.recurseSubmodulesSpacer.setVisible(False)
            ui.recurseSubmodulesTitle.setVisible(False)
            ui.recurseSubmodulesGroupBox.setVisible(False)

        for noRecurseRadio in [ui.createBranchRadioButton, ui.resetHeadRadioButton, ui.mergeRadioButton]:
            noRecurseRadio.toggled.connect(lambda t: ui.recurseSubmodulesGroupBox.setEnabled(not t))

    def bindRadioToOkCaption(self, radio: QRadioButton, okCaption: str, okIcon: str):
        def callback():
            ok = self.ui.buttonBox.button(QDialogButtonBox.StandardButton.Ok)
            ok.setText(okCaption)
            ok.setIcon(stockIcon(okIcon) if okIcon else QIcon())
        radio.clicked.connect(callback)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette import settings
from gitfourchette.forms.aboutdialog import DONATE_URL
from gitfourchette.forms.ui_donateprompt import Ui_DonatePrompt
from gitfourchette.qt import *
from gitfourchette.toolbox import *


class DonatePrompt(QDialog):
    MinStartups = 15
    InitialDelay = 60
    PostponeDelay = 90
    SuggestedAmount = float(3)  # float: for Qt 5 compat (QLocale.toCurrencyString)

    def __init__(self, parent):
        super().__init__(parent)

        self.openDonateLinkOnClose = False

        self.ui = Ui_DonatePrompt()
        self.ui.setupUi(self)
        self.ui.stackedWidget.setCurrentIndex(0)
        self.ui.donateButton.setToolTip(DONATE_URL)

        formatWidgetText(self.ui.label, app=qAppName())
        tweakWidgetFont(self.ui.byeButton, 80)
        tweakWidgetFont(self.ui.thanksLabel, 130, bold=True)

        self.ui.mugshot.setText("")
        self.ui.mugshot.setPixmap(QPixmap("assets:icons/mug"))

        nextDate = QDateTime.fromSecsSinceEpoch(settings.prefs.donatePrompt)
        nextMonth = QLocale().monthName(nextDate.date().month())
        formatWidgetText(self.ui.postponeButton, month=nextMonth)

        amount = QLocale().toCurrencyString(DonatePrompt.SuggestedAmount, "$", 0)
        donateText = formatWidgetText(self.ui.donateButton, amount=amount)
        self.ui.donateButton.setText(donateText + "\n")  # Make room for URL label
        donateButtonLabel = QLabel(f"&nbsp;<br><small><u>{DONATE_URL}</u></small>")
        donateButtonLabel.setAlignment(Qt.AlignmentFlag.AlignCenter)
        donateLayout = QVBoxLayout(self.ui.donateButton)
        donateLayout.setContentsMargins(QMargins())
        donateLayout.addWidget(donateButtonLabel)
        self.ui.donateButton.setCursor(Qt.CursorShape.PointingHandCursor)

        self.ui.donateButton.clicked.connect(self.onDonateButtonPressed)
        self.ui.byeButton.clicked.connect(self.onByeButtonPressed)
        self.ui.postponeButton.clicked.connect(self.close)

        self.finished.connect(self.onDone)

    @staticmethod
    def postpone(days: int) -> int:
        date = QDateTime.currentDateTimeUtc()
        date = date.addDays(days)
        date.setTime(QTime(0, 0, 0))
        scheduledTime = date.toSecsSinceEpoch()

        settings.prefs.donatePrompt = scheduledTime
        settings.prefs.setDirty()
        settings.prefs.write()
        return scheduledTime

    @staticmethod
    def onBoot(parent: QWidget):
        scheduledTime = settings.prefs.donatePrompt

        # Don't show if permanently disabled
        if scheduledTime < 0:
            return

        # Don't show yet if app hasn't been launched enough times
        if settings.history.startups < DonatePrompt.MinStartups:
            return

        # First boot with enough launches, schedule first appearance in the future
        if scheduledTime == 0:
            DonatePrompt.postpone(DonatePrompt.InitialDelay)
            return

        # Don't show yet if scheduled time is still in the future
        now = QDateTime.currentDateTimeUtc()
        if now.toSecsSinceEpoch() < scheduledTime:
            return

        # Don't show yet if any other dialogs are vying for the user's attention on boot
        if parent.findChild(QDialog):
            return

        # Reschedule next appearance
        DonatePrompt.postpone(DonatePrompt.PostponeDelay)

        dp = DonatePrompt(parent)
        dp.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose, True)
        dp.show()

    def onDonateButtonPressed(self):
        settings.prefs.donatePrompt = -1
        settings.prefs.setDirty()
        settings.prefs.write()

        # Show thank-you slide
        self.ui.stackedWidget.setCurrentIndex(1)
        self.setCursor(Qt.CursorShape.WaitCursor)

        self.openDonateLinkOnClose = True
        QTimer.singleShot(1250, self.close)

    def onByeButtonPressed(self):
        settings.prefs.donatePrompt = -1
        settings.prefs.setDirty()
        settings.prefs.write()
        self.close()

    def onDone(self, result):
        if self.openDonateLinkOnClose:
            QDesktopServices.openUrl(QUrl(DONATE_URL))



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import logging
from typing import ClassVar

from gitfourchette.syntax.lexjob import LexJob
from gitfourchette.toolbox.gitutils import shortHash

logger = logging.getLogger(__name__)


class LexJobCache:
    MaxBudget = 1_048_576
    """
    Maximum total size, in bytes, of all the files being lexed.
    This figure sets an upper bound on the total footprint of lexed tokens
    (in the worst case scenario, there would be 1 token per source byte).
    """

    cache: ClassVar[dict[LexJob.KeyType, LexJob]] = {}
    totalFileSize: ClassVar[int] = 0

    @classmethod
    def put(cls, job: LexJob):
        fileKey = job.fileKey

        assert not job.scheduler.isActive()
        assert fileKey not in cls.cache, "LexJob already cached"

        # If the new file is larger than cache capacity, just bail
        if job.fileSize > cls.MaxBudget:
            logger.debug("File too large to fit in cache")
            return

        # Make room in FIFO
        keys = list(cls.cache)
        while cls.totalFileSize + job.fileSize > cls.MaxBudget:
            oldestKey = keys.pop(0)
            cls.evict(oldestKey)

        cls.cache[fileKey] = job
        cls.totalFileSize += job.fileSize
        logger.debug(f"Put {shortHash(fileKey)} (tot: {cls.totalFileSize>>10}K)")

    @classmethod
    def get(cls, fileKey: LexJob.KeyType):
        job = cls.cache.pop(fileKey)
        cls.cache[fileKey] = job  # Bump key
        logger.debug(f"Get {shortHash(fileKey)}")
        return job

    @classmethod
    def evict(cls, fileKey: LexJob.KeyType):
        job = cls.cache.pop(fileKey)
        cls.totalFileSize -= job.fileSize
        logger.debug(f"Del {shortHash(fileKey)} (tot: {cls.totalFileSize>>10:,}K)")
        return job

    @classmethod
    def clear(cls):
        cls.cache.clear()
        cls.totalFileSize = 0



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

"""
Remote access tasks.
"""

import logging
from contextlib import suppress

from gitfourchette.forms.pushdialog import PushDialog
from gitfourchette.forms.textinputdialog import TextInputDialog
from gitfourchette.gitdriver import GitDriver, VanillaFetchStatusFlag, argsIf
from gitfourchette.localization import *
from gitfourchette.nav import NavLocator
from gitfourchette.porcelain import *
from gitfourchette.qt import *
from gitfourchette.tasks import TaskPrereqs, RefreshRepo
from gitfourchette.tasks.branchtasks import MergeBranch
from gitfourchette.tasks.repotask import AbortTask, RepoTask, TaskEffects
from gitfourchette.toolbox import *

logger = logging.getLogger(__name__)


def autoDetectUpstream(repo: Repo, noUpstreamMessage: str = ""):
    branchName = repo.head_branch_shorthand
    branch = repo.branches.local[branchName]

    if not branch.upstream:
        message = noUpstreamMessage or _("Can’t fetch new commits on {0} because this branch isn’t tracking an upstream branch.")
        message = message.format(bquoe(branch.shorthand))
        raise AbortTask(message)

    return branch.upstream


def formatUpdatedRefs(
        updatedTips: dict[str, tuple[str, Oid, Oid]],
        header: str,
        noNewCommits="",
        skipUpToDate=False,
) -> str:
    messages = []
    for ref, (flag, oldTip, newTip) in updatedTips.items():
        rp, rb = RefPrefix.split(ref)
        if not rp:  # no "refs/" prefix, e.g. FETCH_HEAD, etc.
            continue
        if flag == VanillaFetchStatusFlag.UpToDate:
            if skipUpToDate:
                continue
            ps = _("{0} is already up to date with {1}.", tquo(rb), shortHash(oldTip))
        elif flag == VanillaFetchStatusFlag.NewRef:
            ps = _("{0} created: {1}.", tquo(rb), shortHash(newTip))
        elif flag == VanillaFetchStatusFlag.PrunedRef:
            ps = _("{0} deleted, was {1}.", tquo(rb), shortHash(oldTip))
        else:  # ' ' (fast forward), '+' (forced update), '!' (error)
            ps = _("{0}: {1} → {2}.", tquo(rb), shortHash(oldTip), shortHash(newTip))
        messages.append(ps)
    if not messages:
        messages.append(noNewCommits or _("No new commits."))
    return " ".join([header] + messages)


class DeleteRemoteBranch(RepoTask):
    def flow(self, remoteBranchShorthand: str):
        assert not remoteBranchShorthand.startswith(RefPrefix.REMOTES)

        remoteName, branchNameOnRemote = split_remote_branch_shorthand(remoteBranchShorthand)

        text = paragraphs(
            _("Really delete branch {0} from the remote repository?", bquo(remoteBranchShorthand)),
            _("The remote branch will disappear for all users of remote {0}.", bquo(remoteName))
            + " " + _("This cannot be undone!"))
        verb = _("Delete on remote")
        yield from self.flowConfirm(text=text, verb=verb, buttonIcon="SP_DialogDiscardButton")

        self.epilog.effects |= TaskEffects.Remotes | TaskEffects.Refs
        yield from self.flowCallGit(
            "push",
            "--porcelain",
            "--progress",
            "--delete",
            "--",
            remoteName,
            branchNameOnRemote)

        self.epilog.status = _("Remote branch {0} deleted.", tquo(remoteBranchShorthand))


class RenameRemoteBranch(RepoTask):
    def flow(self, remoteBranchShorthand: str):
        assert not remoteBranchShorthand.startswith(RefPrefix.REMOTES)
        remoteName, branchName = split_remote_branch_shorthand(remoteBranchShorthand)
        newBranchName = branchName  # naked name, NOT prefixed with the name of the remote

        reservedNames = self.repo.listall_remote_branches().get(remoteName, [])
        with suppress(ValueError):
            reservedNames.remove(branchName)
        nameTaken = _("This name is already taken by another branch on this remote.")

        dlg = TextInputDialog(
            self.parentWidget(),
            _("Rename remote branch {0}", tquoe(remoteBranchShorthand)),
            _("WARNING: This will rename the branch for all users of the remote!") + "<br>" + _("Enter new name:"))
        dlg.setText(newBranchName)
        dlg.setValidator(lambda name: nameValidationMessage(name, reservedNames, nameTaken))
        dlg.okButton.setText(_("Rename on remote"))

        yield from self.flowDialog(dlg)
        dlg.deleteLater()

        # Naked name, NOT prefixed with the name of the remote
        newBranchName = dlg.lineEdit.text()

        oldShorthand = remoteBranchShorthand

        repo = self.repo
        _remoteName, oldBranchName = split_remote_branch_shorthand(oldShorthand)
        oldRemoteRef = RefPrefix.REMOTES + oldShorthand

        # Find local branches using this upstream
        adjustUpstreams: list[str] = []
        for lb in repo.branches.local:
            with suppress(KeyError):  # KeyError if upstream branch doesn't exist
                if repo.branches.local[lb].upstream_name == oldRemoteRef:
                    adjustUpstreams.append(lb)

        self.epilog.effects |= TaskEffects.Remotes | TaskEffects.Refs

        # First, make a new branch pointing to the same ref as the old one
        refspec1 = f"{RefPrefix.REMOTES}{oldShorthand}:{RefPrefix.HEADS}{newBranchName}"

        # Next, delete the old branch
        refspec2 = f":{RefPrefix.HEADS}{oldBranchName}"

        logger.info(f"Rename remote branch: remote: {remoteName}; refspec: {[refspec1, refspec2]}; "
                    f"adjust upstreams: {adjustUpstreams}")

        # For safety, make sure we're up to date on this branch
        yield from self.flowSubtask(FetchRemoteBranch, oldShorthand)

        # Then go ahead with the push
        yield from self.flowCallGit(
            "push",
            "--porcelain",
            "--progress",
            "--atomic",
            "--",
            remoteName,
            refspec1,
            refspec2)

        new_remote_branch = repo.branches.remote[remoteName + "/" + newBranchName]
        for lb in adjustUpstreams:
            repo.branches.local[lb].upstream = new_remote_branch

        self.epilog.status = _("Remote branch {0} renamed to {1}.", tquo(remoteBranchShorthand), tquo(newBranchName))


class FetchRemotes(RepoTask):
    def flow(self, singleRemoteName: str = ""):
        # Bail now if we don't have any remotes
        if not self.repo.listall_remotes_fast():
            text = paragraphs(
                _("To fetch remote branches, you must first add a remote to your repo."),
                _("You can do so via <i>“Repo &rarr; Add Remote”</i>."))
            raise AbortTask(text)

        self.epilog.effects |= TaskEffects.Remotes | TaskEffects.Refs
        driver = yield from self.flowCallGit(
            "fetch",
            "--prune",
            "--progress",
            *argsIf(GitDriver.supportsFetchPorcelain(), "--porcelain", "--verbose"),
            *argsIf(not singleRemoteName, "--all"),
            *argsIf(bool(singleRemoteName), "--no-all", "--", singleRemoteName))

        # Old git: no --porcelain support, don't attempt to parse the ref table
        if not GitDriver.supportsFetchPorcelain():  # pragma: no cover
            return

        updatedRefs = driver.readFetchPorcelainUpdatedRefs()
        self.epilog.status = formatUpdatedRefs(updatedRefs, _("Fetch complete."), skipUpToDate=True)


class AutoFetchRemotes(RepoTask):
    def isFreelyInterruptible(self) -> bool:
        return True

    def broadcastProcesses(self) -> bool:
        # Don't pop up a ProcessDialog for this task.
        return False

    def flow(self):
        # Bail now if we don't have any remotes
        if not self.repo.listall_remotes_fast():
            return

        self.epilog.effects |= TaskEffects.Remotes | TaskEffects.Refs
        driver = yield from self.flowCallGit(
            "fetch",
            "--all",
            "--prune",
            "--progress",
            *argsIf(GitDriver.supportsFetchPorcelain(), "--porcelain", "--verbose"),
            autoFail=False)

        if driver.exitCode() == 0:
            # Old git: no --porcelain support, don't attempt to parse the ref table
            if not GitDriver.supportsFetchPorcelain():  # pragma: no cover
                return

            updatedRefs = driver.readFetchPorcelainUpdatedRefs()
            self.epilog.status = formatUpdatedRefs(updatedRefs, _("Auto-fetch complete."), skipUpToDate=True)
            return

        # In case of failure, don't steal the user's attention with an obnoxious
        # message box - we may simply be offline and the app may be in the background.
        # Instead, show a subdued message in the status bar.
        message = " ".join([
            _("Couldn’t auto-fetch."),
            _("Git command exited with code {0}.", driver.formatExitCode()),
            tquo(elide(driver.stderrScrollback().strip(), Qt.TextElideMode.ElideRight, 72)),
        ])

        # Prevent chaining RefreshRepo so that the message remains visible in the status bar.
        self.epilog.effects = TaskEffects.Nothing

        raise AbortTask(message, asStatusMessage=True)


class FetchRemoteBranch(RepoTask):
    def flow(self, remoteBranchName: str = "", debrief: bool = True):
        shorthand = remoteBranchName
        if not shorthand:
            upstream = autoDetectUpstream(self.repo)
            shorthand = upstream.shorthand

        remoteName, remoteBranch = split_remote_branch_shorthand(shorthand)
        fullRemoteRef = RefPrefix.REMOTES + shorthand

        self.epilog.effects |= TaskEffects.Remotes | TaskEffects.Refs

        driver = yield from self.flowCallGit(
            "fetch",
            "--progress",
            "--no-tags",
            *argsIf(GitDriver.supportsFetchPorcelain(), "--porcelain", "--verbose"),
            "--",
            remoteName,
            remoteBranch)

        # Old git: no --porcelain support, don't attempt to parse the ref table
        if not GitDriver.supportsFetchPorcelain():  # pragma: no cover
            return

        updatedRefs = driver.readFetchPorcelainUpdatedRefs()
        flag, _oldTarget, newTarget = updatedRefs[fullRemoteRef]

        self.epilog.status = formatUpdatedRefs(
            updatedRefs,
            _("Fetch complete."),
            noNewCommits=_("No new commits on {0}.", lquo(shorthand)),
            skipUpToDate=True)

        # Jump to new commit if there was an update and the branch didn't vanish
        if flag not in [VanillaFetchStatusFlag.UpToDate, VanillaFetchStatusFlag.PrunedRef]:
            self.epilog.jumpTo = NavLocator.inCommit(newTarget)

        if flag == VanillaFetchStatusFlag.PrunedRef:
            # Raise exception to prevent PullBranch from continuing
            # TODO: This does not actually occur when fetching a single branch!
            raise AbortTask(_("{0} has disappeared from the remote server.", bquoe(shorthand)))


class PullBranch(RepoTask):
    def prereqs(self) -> TaskPrereqs:
        return TaskPrereqs.NoUnborn | TaskPrereqs.NoDetached

    def flow(self):
        # Auto-detect the upstream now so we can bail early
        # with a helpful message if there's no upstream.
        noUpstreamMessage = _("Can’t pull new commits into {0} because this branch isn’t tracking an upstream branch.")
        autoDetectUpstream(self.repo, noUpstreamMessage)

        # First, fetch the remote branch.
        # By default, FetchRemoteBranch will fetch the upstream for the current branch.
        yield from self.flowSubtask(FetchRemoteBranch, debrief=False)

        # If we're already up to date, bail now.
        # Note that we're re-resolving the upstream branch after fetching so we have a fresh target.
        upstreamBranch = autoDetectUpstream(self.repo, noUpstreamMessage)
        newUpstreamTarget = upstreamBranch.target
        if self.repo.head_commit_id == newUpstreamTarget:
            self.epilog.status = (
                    _p("toolbar", "Pull") + _(":") + " " +
                    _("Your local branch {0} is already up to date with {1}.",
                      tquo(self.repo.head_branch_shorthand),
                      tquo(upstreamBranch.shorthand)))
            return

        # Let user look at the new state of the graph beneath any dialog boxes.
        yield from self.flowSubtask(RefreshRepo, effectFlags=TaskEffects.Refs, jumpTo=self.epilog.jumpTo)

        # Consume jumpTo (which bubbled up from fetchSubtask) so the next subtask can override it
        self.epilog.jumpTo = NavLocator.Empty

        # Fast-forward, or merge.
        try:
            silentFastForward = self.repo.config.get_bool("pull.ff")
        except (KeyError, GitError):
            silentFastForward = True
        yield from self.flowSubtask(MergeBranch, upstreamBranch.name,
                                    silentFastForward=silentFastForward, autoFastForwardOptionName="pull.ff")


class UpdateSubmodule(RepoTask):
    def flow(self, submoduleName: str):
        submodule = self.repo.submodules[submoduleName]
        self.epilog.effects |= TaskEffects.Workdir
        yield from self.flowCallGit("submodule", "update", "--init", "--", submodule.path)
        self.epilog.status = _("Submodule updated.")


class UpdateSubmodulesRecursive(RepoTask):
    def flow(self):
        self.epilog.effects |= TaskEffects.Workdir
        yield from self.flowCallGit("submodule", "update", "--init", "--recursive")
        self.epilog.status = _("Submodules updated recursively.")


class PushRefspecs(RepoTask):
    def flow(self, remoteName: str, refspecs: list[str]):
        assert remoteName
        assert type(remoteName) is str
        assert type(refspecs) is list

        if remoteName == "*":
            remotes = list(self.repo.remotes)
        else:
            remotes = [self.repo.remotes[remoteName]]

        self.epilog.effects |= TaskEffects.Refs

        for remote in remotes:
            yield from self.flowCallGit("push", "--porcelain", "--progress", "--atomic", remote.name, *refspecs)


class PushBranch(RepoTask):
    def broadcastProcesses(self) -> bool:
        return False

    def flow(self, branchName: str = ""):
        if len(self.repo.remotes) == 0:
            text = paragraphs(
                _("To push a local branch to a remote, you must first add a remote to your repo."),
                _("You can do so via <i>“Repo &rarr; Add Remote”</i>."))
            raise AbortTask(text)

        try:
            if not branchName:
                branchName = self.repo.head_branch_shorthand
            branch = self.repo.branches.local[branchName]
        except (GitError, KeyError) as exc:
            raise AbortTask(_("Please switch to a local branch before performing this action.")) from exc

        dialog = PushDialog(self.repoModel, branch, self.parentWidget())
        dialog.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)

        tryAgain = True
        while tryAgain:
            tryAgain = yield from self.attempt(dialog)

        dialog.accept()

    def attempt(self, dialog: PushDialog):
        # ---------------
        # Show dialog

        yield from self.flowDialog(dialog, proceedSignal=dialog.startOperationButton.clicked)

        # ---------------
        # Perform the push

        command = dialog.buildCommand()
        remoteName = dialog.currentRemoteName

        dialog.setBusy(True)  # Call setBusy *after* buildCommand

        self.epilog.effects |= TaskEffects.Refs
        if "--set-upstream" in command:
            self.epilog.effects |= TaskEffects.Upstreams

        driver = self.createGitProcess(*command)
        dialog.ui.statusForm.connectProcess(driver)
        yield from self.flowStartProcess(driver, autoFail=False)

        gitFailed = driver.exitCode() != 0

        # ---------------
        # Debrief

        # Output format: "<flag> \t <from(local)>:<to(remote)> \t <summary> (<reason>)"
        # But the first and last lines may contain other junk,
        # so skip lines that don't match the pattern (strict=False).
        table = driver.stdoutTable("(.)\t(.+):(.+)\t(.+)", strict=False)

        # Capture summary for the last pushed branch.
        try:
            summary = table[-1][-1]
        except IndexError:
            summary = ""

        dialog.setBusy(False)
        dialog.saveShadowUpstream()

        if not gitFailed:
            # self.epilog.status = RemoteLink.formatUpdatedTipsMessageFromGitOutput(_("Push complete."))
            self.epilog.status = _("Push complete.") + " " + summary
            return gitFailed

        QApplication.beep()
        QApplication.alert(dialog, 500)

        subtitle = ""
        if "[rejected]" in summary:
            subtitle += _("{0} &mdash; The push was rejected.", btag(summary))
        if "(stale info)" in summary:  # Git doesn't provide a hint about this, so add our own
            subtitle += paragraphs(
                _("Your repository’s knowledge of remote branch {0} is out of date. "
                  "The force-push was rejected to prevent data loss.", hquo(dialog.currentRemoteBranchFullName)),
                _("Please fetch remote {0} before pushing again.", hquo(remoteName)))
        errorText = driver.htmlErrorText(subtitle, reformatHintText=True)
        dialog.ui.statusForm.setBlurb(errorText)
        return gitFailed



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import dataclasses
from enum import StrEnum

from gitfourchette.gitdriver.gitdeltafile import GitDeltaFile


class GitConflictSides(StrEnum):
    BothDeleted   = "DD"
    AddedByUs     = "AU"
    AddedByThem   = "UA"
    DeletedByUs   = "DU"
    DeletedByThem = "UD"
    BothAdded     = "AA"
    BothModified  = "UU"

    def hasOurs(self) -> bool:
        return self[0] != "D" and self != GitConflictSides.AddedByThem

    def hasTheirs(self) -> bool:
        return self[1] != "D" and self != GitConflictSides.AddedByUs


@dataclasses.dataclass
class GitConflict:
    sides: GitConflictSides
    ancestor: GitDeltaFile
    ours: GitDeltaFile
    theirs: GitDeltaFile



# Adapted from https://stackoverflow.com/a/68092991
# New: per-line elision in multiline strings

from gitfourchette.qt import *


class QElidedLabel(QLabel):
    _elideMode = Qt.TextElideMode.ElideRight

    def elideMode(self):
        return self._elideMode

    def setElideMode(self, mode):
        if self._elideMode != mode and mode != Qt.TextElideMode.ElideNone:
            self._elideMode = mode
            self.updateGeometry()

    def minimumSizeHint(self):
        return self.sizeHint()

    def sizeHint(self):
        hint = self.fontMetrics().boundingRect(self.text()).size()
        cm = self.contentsMargins()
        margin = self.margin() * 2
        return QSize(
            min(100, hint.width()) + cm.left() + cm.right() + margin,
            min(self.fontMetrics().height(), hint.height()) + cm.top() + cm.bottom() + margin
        )

    def paintEvent(self, event):
        qp = QPainter(self)
        opt = QStyleOptionFrame()
        self.initStyleOption(opt)
        self.style().drawControl(QStyle.ControlElement.CE_ShapedFrame, opt, qp, self)

        elideMode = self.elideMode()
        metrics = self.fontMetrics()
        margin = self.margin()
        m = int(metrics.horizontalAdvance('x') / 2 - margin)  # int() for PyQt5 compat
        r = self.contentsRect().adjusted(margin + m,  margin, -(margin + m), -margin)
        width = r.width()

        elidedText = "\n".join(
            metrics.elidedText(line, elideMode, width)
            for line in self.text().splitlines())

        qp.drawText(r, self.alignment(), elidedText)
        qp.end()



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from __future__ import annotations

import dataclasses
import enum
import re
import traceback
from pathlib import Path

from gitfourchette import settings
from gitfourchette.appconsts import *
from gitfourchette.localization import _, _p
from gitfourchette.nav import NavContext
from gitfourchette.porcelain import RefPrefix, split_remote_branch_shorthand
from gitfourchette.sidebar.sidebarmodel import SidebarItem
from gitfourchette.toolbox import escamp
from gitfourchette.exttools.toolcommands import ToolCommands

if TYPE_CHECKING:
    from gitfourchette.repowidget import RepoWidget


@dataclasses.dataclass
class UserCommand:
    LeaderKey = "Ctrl+K"
    AcceleratorPattern = re.compile(r"(?<![^&]&)&([^&])")
    SeparatorDash = "-"
    AlwaysConfirmPrefix = "?"

    class Token(enum.StrEnum):
        Commit          = "$COMMIT"
        File            = "$FILE"
        FileAbs         = "$FILEABS"
        FileDir         = "$FILEDIR"
        FileDirAbs      = "$FILEDIRABS"
        Head            = "$HEAD"
        HeadBranch      = "$HEADBRANCH"
        HeadUpstream    = "$HEADUPSTREAM"
        Ref             = "$REF"
        Remote          = "$REMOTE"
        Workdir         = "$WORKDIR"

    class MultiTokenError(ValueError):
        tokenErrors: dict[str, Exception]

        def __init__(self, errors):
            self.tokenErrors = errors

    command: str
    userTitle: str = ""
    placeholderTokens: set[str] = dataclasses.field(default_factory=set)
    isSeparator: bool = False
    shortcut: str = ""
    alwaysConfirm: bool = False

    def menuTitle(self) -> str:
        if not self.userTitle:
            title = escamp(self.command)
        else:
            # Don't escamp the user-defined title!
            # Let them define their own accelerator keys.
            title = self.userTitle

        # Qt interprets tabs as separators for keyboard shortcuts.
        # QTextEdit may insert tabs in commands, so override those.
        title = title.replace('\t', ' ')

        if self.alwaysConfirm or settings.prefs.confirmCommands:
            title += "\u2026"  # Add ellipsis character

        return title

    def menuToolTip(self) -> str:
        return self.command if not self.userTitle else ""

    def matchesContext(self, tokenSet: set[str]):
        if self.isSeparator:
            return False
        return any(t in tokenSet for t in self.placeholderTokens)

    def compile(self, context: RepoWidget) -> list[str]:
        tokens = ToolCommands.splitCommandTokens(self.command)
        placeholders = set(ToolCommands.findPlaceholderTokens(tokens))

        replacements = {}
        errors = {}
        for token in placeholders:
            assert token.startswith("$")
            try:
                try:
                    internalIdentifier = UserCommand.Token(token)
                except ValueError as unknownTokenError:
                    raise KeyError(_("Unknown placeholder token")) from unknownTokenError
                callback = getattr(self, f"eval{internalIdentifier.name}")
                replacements[token] = callback(context)
            except Exception as exc:
                errors[token] = exc
                traceback.print_exc()

        if errors:
            raise UserCommand.MultiTokenError(errors)

        tokens = ToolCommands.injectReplacements(tokens, replacements)
        return tokens

    @classmethod
    def evalWorkdir(cls, context: RepoWidget):
        repo = context.repo
        return repo.workdir

    @classmethod
    def evalCommit(cls, context: RepoWidget) -> str:
        repo = context.repo
        locator = context.navLocator

        if locator.context != NavContext.COMMITTED:
            raise ValueError(_("A commit must be selected in the history"))
        commit = repo[locator.commit]
        return str(commit.short_id)

    @classmethod
    def evalHead(cls, context: RepoWidget) -> str:
        repo = context.repo

        if repo.head_is_unborn:
            raise ValueError(_("HEAD cannot be unborn"))
        commit = repo.head_commit
        return str(commit.short_id)

    @classmethod
    def evalRef(cls, context: RepoWidget) -> str:
        sidebarNode = context.sidebar.selectedNode()

        refKinds = {
            SidebarItem.LocalBranch,
            SidebarItem.RemoteBranch,
            SidebarItem.Tag,
        }

        if sidebarNode.kind not in refKinds:
            raise ValueError(_("A ref (local branch, remote branch, or tag) must be selected in the sidebar"))
        return sidebarNode.data

    @classmethod
    def evalHeadBranch(cls, context: RepoWidget) -> str:
        repo = context.repo
        if repo.head_is_unborn or repo.head_is_detached:
            raise ValueError(_("HEAD cannot be unborn or detached"))
        return repo.head_branch_fullname

    @classmethod
    def evalHeadUpstream(cls, context: RepoWidget) -> str:
        repo = context.repo
        if repo.head_is_unborn or repo.head_is_detached:
            raise ValueError(_("HEAD cannot be unborn or detached"))
        branch = repo.branches.local[repo.head_branch_shorthand]
        try:
            return branch.upstream_name
        except KeyError as exc:
            raise ValueError(_("Current branch has no upstream")) from exc

    @classmethod
    def evalFile(cls, context: RepoWidget) -> str:
        locator = context.navLocator
        if not locator.path:
            raise ValueError(_("A file must be selected"))
        return locator.path

    @classmethod
    def evalFileAbs(cls, context: RepoWidget) -> str:
        repo = context.repo
        return repo.in_workdir(cls.evalFile(context))

    @classmethod
    def evalFileDir(cls, context: RepoWidget) -> str:
        path = Path(cls.evalFile(context))
        return path.parent.as_posix()

    @classmethod
    def evalFileDirAbs(cls, context: RepoWidget) -> str:
        path = Path(cls.evalFileAbs(context))
        return path.parent.as_posix()

    @classmethod
    def evalRemote(cls, context: RepoWidget) -> str:
        node = context.sidebar.selectedNode()
        if node.kind == SidebarItem.Remote:
            return node.data
        elif node.kind == SidebarItem.RemoteBranch:
            _prefix, shorthand = RefPrefix.split(node.data)
            remoteName, _branch = split_remote_branch_shorthand(shorthand)
            return remoteName
        raise ValueError(_("A remote or a remote branch must be selected"))

    @staticmethod
    def tokenHelpTable():
        Token = UserCommand.Token
        relSuffix = " " + _p("relative path", "(relative)")
        absSuffix = " " + _p("absolute path", "(absolute)")
        return {
            Token.Commit        : _("Hash of the selected commit in the history"),
            Token.File          : _("Path to the selected file") + relSuffix,
            Token.FileAbs       : _("Path to the selected file") + absSuffix,
            Token.FileDir       : _("Path to the selected file’s parent directory") + relSuffix,
            Token.FileDirAbs    : _("Path to the selected file’s parent directory") + absSuffix,
            Token.Head          : _("Hash of the HEAD commit"),
            Token.HeadBranch    : _("Ref name of the HEAD branch"),
            Token.HeadUpstream  : _("Ref name of the HEAD branch’s upstream"),
            Token.Ref           : _("Name of the selected ref in the sidebar (local branches, remote branches, tags)"),
            Token.Remote        : _("Name of the selected remote in the sidebar"),
            Token.Workdir       : _("Path to the repository’s working directory") + absSuffix,
        }

    @classmethod
    def parseCommandBlock(cls, commandBlock: str):
        for line in commandBlock.splitlines(keepends=False):
            split = line.split("#", 1)

            if len(split) == 1:
                command = line
                comment = ""
            else:
                command, comment = split

            command = command.strip()
            comment = comment.strip()

            if not command:
                # If the comment is all dashes, that's a separator
                if comment and comment == cls.SeparatorDash * len(comment):
                    yield UserCommand("", isSeparator=True)
                continue

            # Force confirmation prompt?
            alwaysConfirm = command.startswith(cls.AlwaysConfirmPrefix)
            if alwaysConfirm:
                command = command.removeprefix(cls.AlwaysConfirmPrefix).strip()

            # Find placeholder tokens
            tokens = ToolCommands.splitCommandTokens(command)
            placeholders = set(ToolCommands.findPlaceholderTokens(tokens))

            # Find accelerator key and derive shortcut from it
            accelMatch = UserCommand.AcceleratorPattern.search(comment)
            if accelMatch:
                accelKey = accelMatch.group(1)
                shortcut = f"{cls.LeaderKey}, {accelKey}"
            else:
                shortcut = ""

            yield UserCommand(
                command=command,
                userTitle=comment,
                placeholderTokens=placeholders,
                shortcut=shortcut,
                alwaysConfirm=alwaysConfirm)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import logging

from gitfourchette import settings
from gitfourchette.codeview.codewindow import CodeWindow
from gitfourchette.diffview.diffdocument import DiffDocument
from gitfourchette.forms.repostub import RepoStub
from gitfourchette.gitdriver import GitDelta, GitDeltaFile, GitStatus, GitConflict, GitDriver
from gitfourchette.gitdriver.lfspointer import LfsObjectCacheMissingError
from gitfourchette.gitdriver.parsers import parseAheadBehind
from gitfourchette.syntax.lexercache import LexerCache
from gitfourchette.syntax.lexjob import LexJob
from gitfourchette.syntax.lexjobcache import LexJobCache
from gitfourchette.diffview.specialdiff import SpecialDiffError, ImageDelta
from gitfourchette.graph import GraphBuildLoop
from gitfourchette.localization import *
from gitfourchette.nav import NavLocator, NavFlags, NavContext
from gitfourchette.porcelain import *
from gitfourchette.qt import *
from gitfourchette.tasks.repotask import RepoTask, TaskEffects, AbortTask
from gitfourchette.toolbox import *

logger = logging.getLogger(__name__)

RENAME_COUNT_THRESHOLD = 100
""" Don't find_similar beyond this number of files in the main diff """

LONG_LINE_THRESHOLD = 10_000

type TAbstractDiffDocument = DiffDocument | GitConflict | ImageDelta | SpecialDiffError


class PrimeRepo(RepoTask):
    """
    This task initializes a RepoModel and a RepoWidget.
    It runs on a RepoStub's RepoTaskRunner, then hands over control to the RepoWidget.
    """

    @classmethod
    def name(cls) -> str:
        return _("Loading repo")

    def flow(self, repoStub: RepoStub):
        # Store repoStub in class instance so that onError() can manipulate it
        # if this coroutine raises an exception
        self.repoStub = repoStub

        from gitfourchette.mainwindow import MainWindow
        from gitfourchette.repowidget import RepoWidget
        from gitfourchette.repomodel import RepoModel
        from gitfourchette.tasks import Jump

        mainWindow = repoStub.window()
        assert isinstance(repoStub, RepoStub)
        assert isinstance(mainWindow, MainWindow)

        mainWindow.statusBar2.showBusyMessage(repoStub.ui.progressLabel.text())

        path = repoStub.workdir
        maxCommits = repoStub.maxCommits

        # Prepare initial locator
        locator = repoStub.locator
        if locator.context == NavContext.SPECIAL:  # Don't attempt to restore a special context
            locator = NavLocator.Empty
        if not locator:  # Fall back to workdir
            locator = NavLocator(NavContext.WORKDIR).withExtraFlags(NavFlags.AllowWriteIndex)

        # Create the repo
        repo = Repo(path, RepositoryOpenFlag.NO_SEARCH)

        if repo.is_bare:
            raise NotImplementedError(_("Sorry, {app} doesn’t support bare repositories.", app=qAppName()))

        # Create RepoModel
        repoModel = RepoModel(repo)
        self.setRepoModel(repoModel)  # required to execute subtasks later

        # Fill in superproject
        with Benchmark("superproject"):
            driver = yield from self.flowCallGit("rev-parse", "--show-superproject-working-tree", workdir=path)
            repoModel.superproject = driver.stdoutScrollback().rstrip()

        # Fill in ahead/behind
        with Benchmark("ahead-behind"):
            driver = yield from self.flowCallGit("for-each-ref", "--format=%(refname:short) %(upstream:track)", "refs/heads")
            repoModel.aheadBehind = dict(parseAheadBehind(driver.stdoutScrollback()))

        # ---------------------------------------------------------------------
        # EXIT UI THREAD
        # ---------------------------------------------------------------------
        yield from self.flowEnterWorkerThread()

        # Get a locale to format numbers on the worker thread
        locale = QLocale()

        # Prime the walker (this might take a while)
        walker = repoModel.primeWalker()

        # Retrieve the number of commits that we loaded last time we opened this repo
        # so we can estimate how long it'll take to load it again
        numCommitsBallpark = settings.history.getRepoNumCommits(repo.workdir)

        if not numCommitsBallpark:
            repoStub.progressFraction.emit(-1.0)  # Indeterminate progress

        # ---------------------------------------------------------------------
        # Build commit sequence

        truncatedHistory = False
        if maxCommits < 0:  # -1 means take maxCommits from prefs. Warning, pref value can be 0, meaning infinity!
            maxCommits = settings.prefs.maxCommits
        if maxCommits == 0:  # 0 means infinity
            maxCommits = 2**63  # ought to be enough
        progressInterval = 1000

        commitSequence = [repoModel.uncommittedChangesMockCommit()]

        for i, commit in enumerate(walker):
            commitSequence.append(commit)

            if i + 1 >= maxCommits or (i + 1 >= progressInterval and repoStub.didAbort):
                truncatedHistory = True
                break

            # Report progress, not too often
            if i % progressInterval == 0 and not repoStub.didAbort:
                message = _("{0} commits…", locale.toString(i))
                repoStub.progressMessage.emit(message)
                if numCommitsBallpark:  # Fill up left half of progress bar
                    repoStub.progressFraction.emit(.5 * min(1.0, i / numCommitsBallpark))
                # Let RepoTaskRunner kill us here (e.g. if closing the tab while we're loading)
                yield from self.flowEnterWorkerThread()

        # Can't abort anymore
        repoStub.progressAbortable.emit(False)

        numCommits = len(commitSequence) - 1
        logger.info(f"{repoModel.shortName}: loaded {numCommits} commits")
        if truncatedHistory:
            message = _("{0} commits loaded (truncated log).", locale.toString(numCommits))
        else:
            message = _("{0} commits total.", locale.toString(numCommits))
            repoStub.progressMessage.emit(message)

        # ---------------------------------------------------------------------
        # Build graph

        def reportGraphProgress(commitNo: int):
            if not numCommits:  # Avoid division by 0
                return
            progress = commitNo / numCommits
            if numCommitsBallpark:
                progress = .5 + .5 * progress  # Fill up the right half of the progress bar.
            self.repoStub.progressFraction.emit(progress)

        hideSeeds = repoModel.getHiddenTips()
        localSeeds = repoModel.getLocalTips()
        buildLoop = GraphBuildLoop(heads=repoModel.getKnownTips(), hideSeeds=hideSeeds, localSeeds=localSeeds)
        buildLoop.onKeyframe = reportGraphProgress
        buildLoop.sendAll(commitSequence)
        repoStub.progressFraction.emit(1.0)

        graph = buildLoop.graph
        repoModel.hiddenCommits = buildLoop.hiddenCommits
        repoModel.foreignCommits = buildLoop.foreignCommits
        repoModel.commitSequence = commitSequence
        repoModel.truncatedHistory = truncatedHistory
        repoModel.graph = graph
        repoModel.hideSeeds = hideSeeds
        repoModel.localSeeds = localSeeds

        # ---------------------------------------------------------------------
        # RETURN TO UI THREAD
        # ---------------------------------------------------------------------
        yield from self.flowEnterUiThread()
        assert not repoStub.didClose, "unexpectedly continuing PrimeRepo after RepoStub was closed"

        # Save commit count (if not truncated)
        if not truncatedHistory:
            settings.history.setRepoNumCommits(repo.workdir, numCommits)

        # Bump repo in history
        settings.history.addRepo(repo.workdir)
        settings.history.setRepoSuperproject(repo.workdir, repoModel.superproject)
        settings.history.write()

        # Finally, prime the UI: Create RepoWidget
        repoStub.taskRunner.repoModel = repoModel
        rw = RepoWidget(repoModel, repoStub.taskRunner, parent=mainWindow)
        # del repoStub.taskRunner

        # Replicate final loading message in status bar
        self.epilog.status = message

        # Jump to initial locator
        repoStub.closing.connect(rw.prepareForDeletion)
        yield from self.flowSubtask(Jump, locator)
        repoStub.closing.disconnect(rw.prepareForDeletion)
        assert not repoStub.didClose, "unexpectedly continuing InstallRepoWidget after RepoStub was closed during Jump"

        rw.refreshNumUncommittedChanges()
        rw.graphView.scrollToRowForLocator(locator, QAbstractItemView.ScrollHint.PositionAtCenter)

        mainWindow.installRepoWidget(rw, mainWindow.tabs.indexOf(repoStub))

        # Once RepoWidget is installed, consider repoStub dead beyond this point
        del repoStub, self.repoStub

        # Refresh tab/window/banner text
        rw.nameChange.emit()

        # It's not necessary to refresh everything again (including workdir patches)
        # after priming the repo.
        self.epilog.effects = TaskEffects.Nothing

        # RepoWidget.refreshRepo may have been called while we were setting up the widget in this task.
        # Clear the stashed effect bits to avoid an unnecessary refresh.
        rw.taskRunner.consumePendingEffectsAndLocator()

        # Focus on some interesting widget within the RepoWidget after loading the repo.
        rw.graphView.setFocus()

    def onError(self, exc: Exception):
        try:
            repoStub = self.repoStub
        except AttributeError:
            pass
        else:
            if not repoStub.didClose:
                repoStub.disableAutoLoad(message=str(exc))

        super().onError(exc)


class LoadPatch(RepoTask):
    def canKill(self, task: RepoTask):
        return isinstance(task, LoadPatch)

    def flow(self, delta: GitDelta, locator: NavLocator):
        try:
            diff = yield from self._getPatch(delta, locator)
        except LfsObjectCacheMissingError as lfsMissing:
            diff = SpecialDiffError.missingLfsObjects(lfsMissing)
        except Exception as exc:
            # Yikes! Don't prevent loading a repo
            summary, details = excStrings(exc)
            diff = SpecialDiffError(escape(summary), icon="SP_MessageBoxCritical", preformatted=details)

        assert onAppThread()

        if isinstance(diff, DiffDocument):
            # Complete existing delta with actual blob hashes.
            # WARNING: GitDelta.id represents a *vanilla* blob, not an LFS blob,
            # so don't update the id if we're diffing an unpacked LFS object.
            if not delta.old.lfs:
                assert diff.oldHash == delta.old.id
            if not delta.new.lfs and not delta.new.isIdValid():
                delta.new.id = diff.newHash

            # Prime lexer
            diff.oldLexJob, diff.newLexJob = self._primeLexJobs(delta)

        return diff

    def _getPatch(
            self,
            delta: GitDelta,
            locator: NavLocator,
    ) -> RepoTask.Flow[TAbstractDiffDocument]:
        if delta.conflict is not None:
            return delta.conflict

        if delta.isSubtreeCommitPatch():
            return SpecialDiffError.submoduleDiff(self.repo, delta, locator)

        if delta.similarity == 100:
            return SpecialDiffError.noChange(self.repo, delta)

        # Special formatting for TYPECHANGE.
        if delta.status == GitStatus.TypeChanged:
            return SpecialDiffError.typeChange(delta)

        # ---------------------------------------------------------------------
        # Load the patch

        # See if we should load an LFS object
        loadLfs = False
        if settings.prefs.lfsAware:
            delta.cacheLfsPointers(self.repo)
            hasOldLfs = bool(delta.old.lfs)
            hasNewLfs = bool(delta.new.lfs)

            if delta.status == GitStatus.Deleted:
                loadLfs = hasOldLfs
            elif delta.status.isAddedOrUntracked:
                loadLfs = hasNewLfs
            else:
                loadLfs = hasOldLfs and hasNewLfs

        # Determine size/line limits
        likelyImage = isImageFormatSupported(delta.old.path) and isImageFormatSupported(delta.new.path)
        if locator.hasFlags(NavFlags.AllowLargeFiles):
            maxLineLength = 0
            maxFileSize = 0
        else:
            maxLineLength = LONG_LINE_THRESHOLD
            if likelyImage:
                maxFileSize = settings.prefs.imageFileThresholdKB * 1024
            else:
                maxFileSize = settings.prefs.largeFileThresholdKB * 1024

        # Don't load large diffs
        if maxFileSize != 0 and GitDeltaFile.SupportsFastSizeBallpark:
            oldBallpark = delta.old.sizeBallpark(self.repo)
            newBallpark = delta.new.sizeBallpark(self.repo)
            ballpark = max(oldBallpark, newBallpark)
            if maxFileSize < ballpark:
                return SpecialDiffError.fileTooLarge(ballpark, maxFileSize, locator, image=likelyImage)

        # If either side is LFS, but not both, see if the contents have changed
        # during the conversion. Performance note: this will cause the entire
        # file to be read, but we're below the large file threshold here.
        if not loadLfs and bool(delta.old.lfs) != bool(delta.new.lfs):
            return SpecialDiffError.lfsMigration(self.repo, delta)

        # Render SVG file if user wants to.
        if (settings.prefs.renderSvg
                and delta.new.path.lower().endswith(".svg")
                and isImageFormatSupported("file.svg")):
            return SpecialDiffError.binaryDiff(self.repo, delta, locator)

        # Build diff command
        if loadLfs:
            tokens = GitDriver.buildDiffCommandLFS(delta)
        else:
            tokens = GitDriver.buildDiffCommand(delta, binary=False, forDisplay=True)

        # Run diff command
        driver = yield from self.flowCallGit(*tokens, autoFail=False)
        patch = driver.stdoutScrollback()

        # Don't display large diffs (legacy pygit2 version)
        # TODO: Remove this once we drop support for pygit2 <= 1.19.1
        if not GitDeltaFile.SupportsFastSizeBallpark and 0 < maxFileSize < len(patch):  # pragma: no cover (old pygit2)
            return SpecialDiffError.fileTooLarge(len(patch), maxFileSize, locator, image=False, legacy=True)

        # Building the diff document on the background thread lets the user
        # interrupt the task, e.g. if dragging the mouse across many commits.
        yield from self.flowEnterWorkerThread()

        # Special case for unstaged files: Before loading the patch, update the
        # GitDelta with fresh filesystem stats (st_mtime_ns). This allows
        # bypassing LoadPatch in a future refresh of the UI if the file isn't
        # modified. (We don't need to stat non-unstaged files because blob
        # hashes are known in advance, so for those, we can simply compare the
        # hashes stored in GitDelta to figure out if it's fresh.)
        if locator.context == NavContext.UNSTAGED:
            delta.new.diskStat = delta.new.stat(self.repo)

        try:
            diff = DiffDocument.fromPatch(patch, maxLineLength)
            diff.document.moveToThread(QApplication.instance().thread())
            return diff
        except DiffDocument.BinaryError:
            return SpecialDiffError.binaryDiff(self.repo, delta, locator)
        except DiffDocument.NoChangeError:
            stderr = driver.stderrScrollback()
            return SpecialDiffError.noChange(self.repo, delta, stderr)
        except DiffDocument.VeryLongLinesError:
            loadAnywayLoc = locator.withExtraFlags(NavFlags.AllowLargeFiles)
            return SpecialDiffError(
                _("This file contains very long lines."),
                linkify(_("[Load diff anyway] (this may take a moment)"), loadAnywayLoc.url()),
                "SP_MessageBoxWarning")
        finally:
            # Exit background thread
            yield from self.flowEnterUiThread()

    def _primeLexJobs(self, delta: GitDelta) -> tuple[LexJob | None, LexJob | None]:
        assert onAppThread()

        if not settings.prefs.isSyntaxHighlightingEnabled():
            return None, None

        lexer = LexerCache.getLexerFromPath(delta.new.path, settings.prefs.pygmentsPlugins)

        if lexer is None:
            return None, None

        # Avoid awkward syntax highlighting of LFS pointers
        # when diffing against a vanilla Git blob
        if (not delta.old.isId0() and not delta.new.isId0()
                and bool(delta.old.lfs) != bool(delta.new.lfs)):
            return None, None

        oldLexJob = self._primeSingleLexJob(lexer, delta.old)
        newLexJob = self._primeSingleLexJob(lexer, delta.new)
        return oldLexJob, newLexJob

    def _primeSingleLexJob(self, lexer, file: GitDeltaFile):
        if file.isId0() or file.isEmptyBlob():
            return None

        # Figure out a key to look up this file in the lex cache.
        if file.isIdValid():
            # Use blob SHA-1 (most common case)
            key = file.id
        elif file.hasDiskStat():
            # Blob SHA-1 not available, but we've got a stat
            # (E.g. unstaged modification to an LFS file)
            mtime, size = file.diskStat
            key = f"disk:{mtime},{size}"
        else:
            raise NotImplementedError("need valid blob id or stat for lexing")

        try:
            return LexJobCache.get(key)
        except KeyError:
            pass

        data = file.read(self.repo)
        job = LexJob(lexer, data, key)

        assert job.fileKey == key
        assert job.fileKey not in LexJobCache.cache
        LexJobCache.put(job)

        return job


class LoadPatchInNewWindow(RepoTask):
    def flow(self, delta: GitDelta, locator: NavLocator):
        if CodeWindow.activateExistingWindow(locator):
            return

        diffDocument = yield from self.flowSubtask(LoadPatch, delta, locator)

        if not isinstance(diffDocument, DiffDocument):
            raise AbortTask(_("Only text diffs may be opened in a separate window."), icon="information")

        from gitfourchette.diffview.diffview import DiffView

        assert locator.context == NavContext.COMMITTED
        title = f"{locator.path} @ {shortHash(locator.commit)}"

        diffWindow = CodeWindow(DiffView, locator)
        diffWindow.setObjectName("DetachedDiffWindow")
        diffWindow.setWindowTitle(title)
        diff = diffWindow.codeView
        assert isinstance(diff, DiffView)
        diff.replaceDocument(self.repo, delta, locator, diffDocument)
        diffWindow.show()

        self.rw.aboutToDelete.connect(diffWindow.close)


class DownloadLfsObjects(RepoTask):
    def flow(self, errors: LfsObjectCacheMissingError, informativeName: str = ""):
        for pointer in errors.pointers:
            # If successful, `git lfs smudge` will cache the LFS object in .git/lfs/objects
            # Note: the name isn't significant and is only for progress reporting
            driver = self.createGitProcess("lfs", "smudge", informativeName or pointer.id[:7])

            # Discard stdout
            driver.setStandardOutputFile(QProcess.nullDevice())

            yield from self.flowStartProcess(driver, stdin=pointer.pointerText)
            self.epilog.effects |= TaskEffects.DefaultRefresh



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import glob
import re
from pathlib import Path

from gitfourchette import pycompat  # noqa: F401 - glob.translate for Python 3.12
from gitfourchette.forms.brandeddialog import convertToBrandedDialog
from gitfourchette.forms.ui_ignorepatterndialog import Ui_IgnorePatternDialog
from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.toolbox import *


class IgnorePatternDialog(QDialog):
    def __init__(self, path: str, parent: QWidget):
        super().__init__(parent)

        ui = Ui_IgnorePatternDialog()
        ui.setupUi(self)
        self.ui = ui

        self.seedPath = path

        for k, v in self._samplePatternOptions(path).items():
            ui.patternEdit.addItemWithPreview(k, k, v)
        ui.patternEdit.setEditText("/" + path)

        for patternFile, patternFileDescription in self._excludePathOptions().items():
            ui.fileEdit.addItemWithPreview(patternFile, patternFile, patternFileDescription)

        validator = ValidatorMultiplexer(self)
        validator.connectInput(ui.patternEdit.lineEdit(), self.validate)
        validator.run(silenceEmptyWarnings=True)

        subtitle = _("Based off: {0}", hquo(path))
        convertToBrandedDialog(self, subtitleText=subtitle)

    @property
    def pattern(self) -> str:
        return self.ui.patternEdit.currentText()

    @property
    def excludePath(self) -> str:
        return self.ui.fileEdit.currentData(Qt.ItemDataRole.UserRole)

    def validate(self, pattern: str):
        if not pattern:
            return _("Please fill in this field.")

        seedPath = self.seedPath

        # Doctor the pattern to match glob rules
        if "/" not in pattern:
            pattern = "**/" + pattern
        else:
            pattern = pattern.removeprefix("/")

        regex1 = glob.translate(pattern, recursive=True, include_hidden=True)
        regex2 = glob.translate(pattern.removesuffix("/") + "/**", recursive=True, include_hidden=True)

        if re.match(regex1, seedPath):
            return ""

        if re.match(regex2, seedPath):
            return ""

        return _("This pattern doesn’t match the file that you selected.")

    @staticmethod
    def _samplePatternOptions(rawPath: str):
        path = Path(rawPath)

        samples = {
            "/" + rawPath: _("Just this specific path"),
        }

        # pd = str(pp.parent)
        absParent = "/" if str(path.parent) == "." else f"/{path.parent}/"
        if absParent != "/":
            samples[absParent] = _("Everything in this directory")

        suffix = path.suffix
        if suffix:
            wildSuffix = f"*{suffix}"
            samples[f"{absParent}**/{wildSuffix}"] = _("All {0} files in this directory tree", wildSuffix)
            samples[wildSuffix] = _("Any {0} file", wildSuffix)

        samples[path.name] = _("Any file with this exact name")

        return samples

    @staticmethod
    def _excludePathOptions():
        return {
            ".gitignore": _("Share pattern with other contributors"),
            ".git/info/exclude": _("Keep pattern private on this machine")
        }



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.blameview.blamemodel import BlameModel
from gitfourchette.graphview.commitlogmodel import CommitLogModel, SpecialRow
from gitfourchette.localization import _
from gitfourchette.qt import *
from gitfourchette.repomodel import UC_FAKEID
from gitfourchette.toolbox import onAppThread


class BlameScrubberModel(QAbstractListModel):
    blameModel: BlameModel

    def __init__(self, blameModel: BlameModel, parent: QWidget):
        self.blameModel = blameModel

        super().__init__(parent)

        if APP_DEBUG and HAS_QTEST:
            self.modelTester = QAbstractItemModelTester(self)

    def columnCount(self, parent = ...):
        return 1

    def rowCount(self, parent = ...) -> int:
        return len(self.blameModel.revList.sequence)

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole):
        assert index.isValid()

        row = index.row()
        revision = self.blameModel.revList.sequence[row]

        if APP_TESTMODE and role == Qt.ItemDataRole.DisplayRole:
            # DisplayRole is intended as a unit test helper only.
            # BlameScrubberDelegate doesn't render this role.
            if revision.commitId == UC_FAKEID:  # for unit tests
                return _("Uncommitted Changes in Working Directory")
            return self.blameModel.repo.peel_commit(revision.commitId).message

        elif role == CommitLogModel.Role.Commit:
            assert onAppThread()
            try:
                return self.blameModel.repo.peel_commit(revision.commitId)
            except KeyError:
                return None

        elif role == CommitLogModel.Role.Oid:
            return revision.commitId

        elif role == CommitLogModel.Role.SpecialRow:
            return SpecialRow.UncommittedChanges if revision.commitId == UC_FAKEID else SpecialRow.Commit

        elif role == CommitLogModel.Role.BlameRevision:
            return revision

        return None



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import logging
import os.path

from gitfourchette import settings
from gitfourchette.forms.ui_repostub import Ui_RepoStub
from gitfourchette.localization import *
from gitfourchette.nav import NavLocator
from gitfourchette.qt import *
from gitfourchette.toolbox import *

_logger = logging.getLogger(__name__)


class RepoStub(QWidget):
    """
    Placeholder for RepoWidget shown when RepoModel isn't ready.
    """

    # Signals for cross-thread control from PrimeRepo task
    progressFraction = Signal(float)
    progressMessage = Signal(str)
    progressAbortable = Signal(bool)
    closing = Signal()

    def __init__(self, parent=None, workdir="", locator=NavLocator.Empty, maxCommits=-1):
        super().__init__(parent)

        from gitfourchette.tasks import RepoTaskRunner

        self.workdir = os.path.normpath(workdir)
        self.locator = locator
        self.maxCommits = maxCommits
        self.didAbort = False
        self.didClose = False

        self.ui = Ui_RepoStub()
        self.ui.setupUi(self)
        self.setObjectName(f"{self.__class__.__name__}({self.getTitle()})")  # override setupUi's name
        self.setWindowTitle(self.getTitle())

        # Connect buttons
        self.ui.promptLoadButton.clicked.connect(self.loadNow)
        self.ui.progressAbortButton.clicked.connect(self.onAbortButtonClicked)

        self.progressFraction.connect(self.setProgress)
        self.progressMessage.connect(self.ui.progressLabel.setText)
        self.progressAbortable.connect(self.onChangeProgressInterruptable)
        self.ui.progressPage.setCursor(Qt.CursorShape.BusyCursor)

        tweakWidgetFont(self.ui.promptNameLabel, bold=True)
        self.ui.promptNameLabel.setText(tquo(self.getTitle()))
        self.ui.promptIcon.setPixmap(stockIcon("image-missing").pixmap(96))
        self.ui.progressAbortButton.setIcon(stockIcon("SP_BrowserStop"))

        # Initialize labels, etc.
        self.resetUi()

        # We need a task runner to run the initialization task (PrimeRepo).
        self.taskRunner = RepoTaskRunner(parent=self)
        self.taskRunner.setObjectName(f"RepoTaskRunner(STUB: {self.getTitle()})")

    def close(self) -> bool:
        assert not self.didClose, "RepoStub is already closing!"

        self.didAbort = True
        self.didClose = True

        self.ui.progressLabel.setText(_("Aborting. Just a moment…"))
        self.ui.progressAbortButton.setEnabled(False)
        self.setProgress(-1)

        if self.taskRunner.isBusy():
            _logger.debug(f"Interrupting priming task because {self.objectName()} was force closed!")
            self.taskRunner.killCurrentTask()
            self.taskRunner.joinKilledTask()

        self.closing.emit()
        return super().close()

    def resetUi(self):
        self.didAbort = False
        self.ui.retranslateUi(self)
        self.ui.promptIcon.setText("")
        self.ui.promptIcon.setVisible(False)
        self.ui.progressLabel.setText(_("Opening {0}…", tquo(self.getTitle())))
        self.ui.stackedWidget.setCurrentWidget(self.ui.progressPage)
        self.ui.progressAbortButton.setEnabled(True)
        self.setProgress(0)

    def disableAutoLoad(self, message=""):
        self.ui.stackedWidget.setCurrentWidget(self.ui.promptPage)

        if message:
            self.ui.promptReadyLabel.setText(message)
            self.ui.promptIcon.setVisible(True)
            self.ui.promptLoadButton.setText(_("Try to reload"))

    def willAutoLoad(self):
        return self.ui.stackedWidget.currentWidget() is self.ui.progressPage

    def onAbortButtonClicked(self):
        self.didAbort = True
        self.ui.progressLabel.setText(_("Loading interrupted. Just a moment…"))
        self.ui.progressAbortButton.setEnabled(False)

    def setProgress(self, progress: float):
        if progress < 0:
            self.ui.progressBar.setRange(0, 0)
            self.ui.progressBar.setValue(0)
        else:
            self.ui.progressBar.setRange(0, 1000)
            self.ui.progressBar.setValue(int(progress * 1000))

    def onChangeProgressInterruptable(self, abortable: bool):
        self.ui.progressAbortButton.setEnabled(abortable and self.didAbort)

    def overridePendingLocator(self, locator: NavLocator):
        self.locator = locator

    def getTitle(self):
        return settings.history.getRepoTabName(self.workdir)

    def superproject(self):
        return settings.history.getRepoSuperproject(self.workdir)

    def isPriming(self):
        from gitfourchette.tasks.loadtasks import PrimeRepo
        task = self.taskRunner.currentTask
        priming = isinstance(task, PrimeRepo)
        assert not task or priming
        return priming

    def loadNow(self):
        assert not self.isPriming(), "attempting to load RepoStub twice"
        from gitfourchette.tasks.loadtasks import PrimeRepo
        self.resetUi()
        PrimeRepo.invoke(self, repoStub=self)



# Form implementation generated from reading ui file 'donateprompt.ui'
#
# Created by: PyQt6 UI code generator 6.10.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_DonatePrompt(object):
    def setupUi(self, DonatePrompt):
        DonatePrompt.setObjectName("DonatePrompt")
        DonatePrompt.resize(475, 360)
        DonatePrompt.setWindowTitle("")
        self.verticalLayout_2 = QVBoxLayout(DonatePrompt)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        spacerItem = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.verticalLayout_2.addItem(spacerItem)
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setSpacing(20)
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem1 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.label = QLabel(parent=DonatePrompt)
        self.label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.label.setWordWrap(True)
        self.label.setTextInteractionFlags(Qt.TextInteractionFlag.NoTextInteraction)
        self.label.setObjectName("label")
        self.horizontalLayout.addWidget(self.label)
        self.mugshot = QLabel(parent=DonatePrompt)
        self.mugshot.setText("MUG")
        self.mugshot.setTextInteractionFlags(Qt.TextInteractionFlag.NoTextInteraction)
        self.mugshot.setObjectName("mugshot")
        self.horizontalLayout.addWidget(self.mugshot)
        spacerItem2 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.horizontalLayout.addItem(spacerItem2)
        self.verticalLayout_2.addLayout(self.horizontalLayout)
        spacerItem3 = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.verticalLayout_2.addItem(spacerItem3)
        self.stackedWidget = QStackedWidget(parent=DonatePrompt)
        self.stackedWidget.setObjectName("stackedWidget")
        self.page = QWidget()
        self.page.setObjectName("page")
        self.gridLayout_2 = QGridLayout(self.page)
        self.gridLayout_2.setObjectName("gridLayout_2")
        spacerItem4 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.gridLayout_2.addItem(spacerItem4, 3, 4, 1, 1)
        self.donateButton = QPushButton(parent=self.page)
        self.donateButton.setDefault(True)
        self.donateButton.setObjectName("donateButton")
        self.gridLayout_2.addWidget(self.donateButton, 1, 1, 1, 3)
        self.byeButton = QPushButton(parent=self.page)
        self.byeButton.setFlat(True)
        self.byeButton.setObjectName("byeButton")
        self.gridLayout_2.addWidget(self.byeButton, 3, 2, 1, 1)
        spacerItem5 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.gridLayout_2.addItem(spacerItem5, 3, 1, 1, 1)
        spacerItem6 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.gridLayout_2.addItem(spacerItem6, 3, 3, 1, 1)
        spacerItem7 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.gridLayout_2.addItem(spacerItem7, 3, 0, 1, 1)
        self.postponeButton = QPushButton(parent=self.page)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.postponeButton.sizePolicy().hasHeightForWidth())
        self.postponeButton.setSizePolicy(sizePolicy)
        self.postponeButton.setObjectName("postponeButton")
        self.gridLayout_2.addWidget(self.postponeButton, 2, 1, 1, 3)
        self.stackedWidget.addWidget(self.page)
        self.page_2 = QWidget()
        self.page_2.setObjectName("page_2")
        self.verticalLayout = QVBoxLayout(self.page_2)
        self.verticalLayout.setObjectName("verticalLayout")
        self.thanksLabel = QLabel(parent=self.page_2)
        self.thanksLabel.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.thanksLabel.setObjectName("thanksLabel")
        self.verticalLayout.addWidget(self.thanksLabel)
        self.stackedWidget.addWidget(self.page_2)
        self.verticalLayout_2.addWidget(self.stackedWidget)

        self.retranslateUi(DonatePrompt)
        self.stackedWidget.setCurrentIndex(0)
        DonatePrompt.setTabOrder(self.donateButton, self.postponeButton)
        DonatePrompt.setTabOrder(self.postponeButton, self.byeButton)

    def retranslateUi(self, DonatePrompt):
        self.label.setText(_p("DonatePrompt", "Hi! I hope you’re enjoying {app}.\n"
"\n"
"If {app} helps you get work done, please consider making a small donation! This will encourage me to keep developing it. Thank you!\n"
"\n"
"-Iliyas"))
        self.donateButton.setText(_p("DonatePrompt", "Donate {amount}"))
        self.byeButton.setText(_p("DonatePrompt", "Never show this again"))
        self.postponeButton.setText(_p("DonatePrompt", "Remind me in {month}"))
        self.thanksLabel.setText(_p("DonatePrompt", "THANK YOU!"))



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette import trtables
from gitfourchette import settings
from gitfourchette.application import GFApplication
from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.settings import WhitespaceMode
from gitfourchette.toolbox import *


class DiffButtons(QWidget):
    def __init__(self, parent: QWidget):
        super().__init__(parent)

        self.diffMethodActions: dict[WhitespaceMode, QAction] = {}

        self.contextButton = self._makeContextLinesButton()
        self.wordWrapButton = self._makeToggle("diff-wrap", "wordWrap")
        self.showWhitespaceButton = self._makeToggle("diff-show-whitespace", "showWhitespace")
        self.whitespaceModeButton = self._makeWhitespaceDiffButton()
        self.svgButton = self._makeToggle("diff-svg", "renderSvg")
        self.svgButton.setToolTip(_("SVG image preview"))

        self.buttons = [
            self.svgButton,
            self.contextButton,
            self.wordWrapButton,
            self.showWhitespaceButton,
            self.whitespaceModeButton,
        ]

        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 2, 0)
        layout.setSpacing(2)
        for button in self.buttons:
            layout.addWidget(button)

    # -------------------------------------------------------------------------
    # Constructor helpers

    def _makeWhitespaceDiffButton(self):
        button = QToolButton(self)
        menu = QMenu(button)

        actionGroup = QActionGroup(menu)
        actionGroup.setExclusive(True)

        for mode in WhitespaceMode:
            label = escamp(trtables.enum(mode))
            iconName = f"diff-whitespace-{mode or 'strict'}"
            action = QAction(stockIcon(iconName), label, button)
            action.triggered.connect(lambda _dummy, m=mode: self.setWhitespaceMode(m))
            action.setActionGroup(actionGroup)
            action.setCheckable(True)
            menu.addAction(action)
            self.diffMethodActions[mode] = action

        menu.insertSeparator(list(self.diffMethodActions.values())[1])

        button.setMenu(menu)
        button.setPopupMode(QToolButton.ToolButtonPopupMode.InstantPopup)
        button.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonIconOnly)

        # We'll "press" the button when the whitespace mode is anything but Strict.
        button.setCheckable(True)

        return button

    def _makeContextLinesButton(self):
        button = QToolButton(self)
        menu = QMenu(button)

        actionGroup = QActionGroup(menu)
        actionGroup.setExclusive(True)

        container = QWidget()
        layout = QHBoxLayout(container)

        spinbox = QSpinBox()
        spinbox.setRange(0, 32)  # TODO: couple with PrefsDialog bounds
        spinbox.valueChanged.connect(self.setContextLines)
        spinbox.lineEdit().setAlignment(Qt.AlignmentFlag.AlignCenter)

        t1, t2 = _("Show up to # context lines").split("#")
        layout.addWidget(QLabel(t1))
        layout.addWidget(spinbox)
        layout.addWidget(QLabel(t2))

        def aboutToShowContextLinesMenu():
            spinbox.setValue(settings.prefs.contextLines)
            spinbox.setFocus()
            spinbox.selectAll()

        widgetAction = QWidgetAction(menu)
        widgetAction.setDefaultWidget(container)
        menu.addAction(widgetAction)
        menu.aboutToShow.connect(aboutToShowContextLinesMenu)

        button.setToolTip(_("Context lines"))
        button.setMenu(menu)
        button.setPopupMode(QToolButton.ToolButtonPopupMode.InstantPopup)
        button.setText(_("Context"))
        button.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextBesideIcon)

        return button

    def _makeToggle(self, icon: str, prefKey: str):
        button = QToolButton(self)
        button.setCheckable(True)
        button.setIcon(stockIcon(icon))
        button.setToolTip(trtables.prefKey(prefKey))
        button.toggled.connect(lambda checked: GFApplication.applyPrefs(**{prefKey: checked}))
        return button

    # -------------------------------------------------------------------------
    # Sync with preferences

    def refreshPrefs(self):
        with QSignalBlockerContext(
                *self.buttons,
                *self.diffMethodActions.values(),
        ):
            self.wordWrapButton.setChecked(settings.prefs.wordWrap)
            self.showWhitespaceButton.setChecked(settings.prefs.showWhitespace)
            self.contextButton.setIcon(stockIcon("diff-context-lines", f"$TEXT$={settings.prefs.contextLines}"))

            mode = settings.prefs.whitespaceMode
            for m, action in self.diffMethodActions.items():
                action.setChecked(m == mode)
            action = self.diffMethodActions[mode]

            toolTip = stripAccelerators(action.text())
            self.whitespaceModeButton.setIcon(action.icon())
            self.whitespaceModeButton.setToolTip(toolTip)

            # "Press" the button when the mode is anything but Strict.
            self.whitespaceModeButton.setChecked(mode != WhitespaceMode.Strict)

            self.svgButton.setChecked(settings.prefs.renderSvg)

    # -------------------------------------------------------------------------
    # Button callbacks

    @staticmethod
    def setWhitespaceMode(mode: WhitespaceMode):
        GFApplication.applyPrefs(whitespaceMode=mode)

    @staticmethod
    def setContextLines(n: int):
        GFApplication.applyPrefs(contextLines=n)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import dataclasses
import enum
import re

from gitfourchette.localization import *

LfsPointerMagic = "version https://git-lfs.github.com/spec/v1\n"
LfsPointerMagicBytes = LfsPointerMagic.encode("utf-8")
LfsPointerPattern = re.compile(rf"^{LfsPointerMagic}oid sha256:([0-9a-f]+)\nsize (\d+)")


class LfsPointerState(enum.IntEnum):
    NoPointer = -1
    """ This file isn't stored as LFS. """

    Unknown = 0
    """ This file's LFS status hasn't been cached yet. """

    Valid = 1
    """ This file is stored as LFS. """

    UnstagedTentative = 2
    """
    This file is expected to be stored as LFS once it's staged.
    It may not have an object file in the LFS database yet.
    """


@dataclasses.dataclass(frozen=True)
class LfsPointer:
    state: LfsPointerState = LfsPointerState.Unknown
    """ Status of this LFS pointer. """

    id: str = ""
    """ SHA-256 hexadecimal hash of the LFS object. """

    size: int = -1
    """ Size of the LFS object. -1 means unknown. """

    objectPath: str = "/dev/null"
    """ Absolute path to the LFS object file. """

    pointerText: str = ""
    """ Raw LFS pointer text (i.e. the contents of the regular Git blob). """

    def __bool__(self) -> bool:
        return self.state > LfsPointerState.Unknown

    def isTentative(self) -> bool:
        return self.state == LfsPointerState.UnstagedTentative


class LfsObjectCacheMissingError(LookupError):
    def __init__(self, *pointers: LfsPointer):
        title = _n("Object missing from local LFS cache.",
                   "Objects missing from local LFS cache.", n=len(pointers))
        super().__init__(title)
        self.pointers = pointers



# Form implementation generated from reading ui file 'welcomewidget.ui'
#
# Created by: PyQt6 UI code generator 6.10.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from gitfourchette.localization import *
from gitfourchette.qt import *


class Ui_WelcomeWidget(object):
    def setupUi(self, WelcomeWidget):
        WelcomeWidget.setObjectName("WelcomeWidget")
        WelcomeWidget.resize(521, 288)
        self.gridLayout = QGridLayout(WelcomeWidget)
        self.gridLayout.setObjectName("gridLayout")
        self.cloneRepoButton = QPushButton(parent=WelcomeWidget)
        self.cloneRepoButton.setObjectName("cloneRepoButton")
        self.gridLayout.addWidget(self.cloneRepoButton, 9, 1, 1, 1)
        self.newRepoButton = QPushButton(parent=WelcomeWidget)
        self.newRepoButton.setObjectName("newRepoButton")
        self.gridLayout.addWidget(self.newRepoButton, 7, 1, 1, 1)
        spacerItem = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.gridLayout.addItem(spacerItem, 8, 0, 1, 1)
        self.openRepoButton = QPushButton(parent=WelcomeWidget)
        self.openRepoButton.setObjectName("openRepoButton")
        self.gridLayout.addWidget(self.openRepoButton, 8, 1, 1, 1)
        spacerItem1 = QSpacerItem(20, 27, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.gridLayout.addItem(spacerItem1, 10, 1, 1, 1)
        spacerItem2 = QSpacerItem(60, 1, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.gridLayout.addItem(spacerItem2, 8, 2, 1, 1)
        self.recentReposButton = QPushButton(parent=WelcomeWidget)
        self.recentReposButton.setObjectName("recentReposButton")
        self.gridLayout.addWidget(self.recentReposButton, 5, 1, 1, 1)
        spacerItem3 = QSpacerItem(20, 27, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.gridLayout.addItem(spacerItem3, 0, 1, 1, 1)
        spacerItem4 = QSpacerItem(20, 27, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        self.gridLayout.addItem(spacerItem4, 4, 1, 1, 1)
        spacerItem5 = QSpacerItem(20, 20, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        self.gridLayout.addItem(spacerItem5, 6, 1, 1, 1)
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setSpacing(12)
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem6 = QSpacerItem(40, 1, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.horizontalLayout.addItem(spacerItem6)
        self.logoLabel = QLabel(parent=WelcomeWidget)
        self.logoLabel.setText("(LOGO)")
        self.logoLabel.setAlignment(Qt.AlignmentFlag.AlignRight|Qt.AlignmentFlag.AlignTrailing|Qt.AlignmentFlag.AlignVCenter)
        self.logoLabel.setObjectName("logoLabel")
        self.horizontalLayout.addWidget(self.logoLabel)
        self.welcomeLabel = QLabel(parent=WelcomeWidget)
        self.welcomeLabel.setTextInteractionFlags(Qt.TextInteractionFlag.NoTextInteraction)
        self.welcomeLabel.setObjectName("welcomeLabel")
        self.horizontalLayout.addWidget(self.welcomeLabel)
        spacerItem7 = QSpacerItem(40, 1, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.horizontalLayout.addItem(spacerItem7)
        self.gridLayout.addLayout(self.horizontalLayout, 1, 0, 1, 3)

        self.retranslateUi(WelcomeWidget)
        WelcomeWidget.setTabOrder(self.recentReposButton, self.newRepoButton)
        WelcomeWidget.setTabOrder(self.newRepoButton, self.openRepoButton)
        WelcomeWidget.setTabOrder(self.openRepoButton, self.cloneRepoButton)

    def retranslateUi(self, WelcomeWidget):
        WelcomeWidget.setWindowTitle(_p("WelcomeWidget", "Welcome"))
        self.cloneRepoButton.setText(_p("WelcomeWidget", "Clone remote repository"))
        self.newRepoButton.setText(_p("WelcomeWidget", "New empty repository"))
        self.openRepoButton.setText(_p("WelcomeWidget", "Open local repository"))
        self.recentReposButton.setText(_p("WelcomeWidget", "Recent repositories"))
        self.welcomeLabel.setText(_p("WelcomeWidget", "Welcome to<br>{app}"))



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from collections.abc import Iterable

from gitfourchette.diffview.diffdocument import LineData
from gitfourchette.gitdriver import GitDelta, GitStatus
from gitfourchette.porcelain import FileMode

QUOTE_PATH_ESCAPES = {
    '"': '\\"',
    '\a': '\\a',
    '\b': '\\b',
    '\t': '\\t',
    '\n': '\\n',
    '\v': '\\v',
    '\f': '\\f',
    '\r': '\\r',
    '\\': '\\\\',
    # Although we're not technically escaping the space character, it's
    # included in this dict to force any paths containing spaces to be quoted.
    ' ': ' ',
}
""" Predefined character escapes for `quotePath`. """


def quotePath(path: str) -> str:
    # If no escaping is needed, we can spit back the input path verbatim.
    verbatim = True

    # Build a safe (quoted + escaped) path.
    safePath = ['"']

    for char in path:
        # See if we should escape this character
        try:
            char = QUOTE_PATH_ESCAPES[char]
            verbatim = False
        except KeyError:
            # This character has no predefined escape.
            # If it's a printable ASCII char, we can copy it verbatim,
            # otherwise it should be encoded as octal-escaped UTF-8.
            isPrintableAscii = 0x21 <= ord(char) <= 0x7e
            if not isPrintableAscii:
                char = "".join(f"\\{byte:03o}" for byte in char.encode("utf-8"))
                verbatim = False

        safePath.append(char)

    if verbatim:
        # None of the characters had to be escaped
        assert path == "".join(safePath).removeprefix('"')
        return path

    safePath.append('"')
    return "".join(safePath)


def getPatchPreamble(delta: GitDelta, reverse=False) -> list[str]:
    old = delta.old
    new = delta.new

    if not reverse:
        # Not reversing. Old/new sides are correct.
        pass
    elif delta.status == GitStatus.Deleted:
        # Reversing lines within a deleted file. Swap old/new sides.
        new, old = old, new
        assert old.isId0()  # 'old' side is now the deleted file
    else:
        # When reversing lines within a patch, stick to the new file
        # to avoid changing the file's name or mode.
        old = new

    oldPathQuoted = quotePath(f"a/{old.path}")
    newPathQuoted = quotePath(f"b/{new.path}")
    preamble = [f"diff --git {oldPathQuoted} {newPathQuoted}\n"]

    oldExists = not old.isId0()
    newExists = not new.isId0()

    if not oldExists:
        preamble.append(f"new file mode {new.mode:06o}\n")
    elif old.mode != new.mode or new.mode != FileMode.BLOB:
        preamble.append(f"old mode {old.mode:06o}\n")
        preamble.append(f"new mode {new.mode:06o}\n")

    # Work around libgit2 bug: if a patch lacks the "index" line,
    # libgit2 will fail to parse it if there are "old mode"/"new mode" lines.
    # Also, even if the patch is successfully parsed as a Diff, and we need to
    # regenerate it (from the Diff), libgit2 may fail to re-create the
    # "---"/"+++" lines and it'll therefore fail to parse its own output.
    preamble.append(f"index {old.id}..{'f' * 40}\n")

    preamble.append(f"--- {oldPathQuoted if oldExists else '/dev/null'}\n")
    preamble.append(f"+++ {newPathQuoted if newExists else '/dev/null'}\n")

    return preamble


def writeContext(subpatch: list[str], reverse: bool, lines: Iterable[LineData]):
    skipOrigin = '-' if reverse else '+'
    for line in lines:
        assert line.origin in " +-", f"unknown origin {line.origin}"

        if line.origin == skipOrigin:
            # Skip that line entirely
            continue

        # Make it a context line
        subpatch.append(" ")
        subpatch.append(line.text)
        subpatch.append(line.hiddenSuffix)


def extractSubpatch(
        masterDelta: GitDelta,
        lines: list[LineData],
        spanStart: int,
        spanEnd: int,
        reverse: bool
) -> str:
    """
    Create a patch (in unified diff format) from a range of selected lines in a diff.
    """

    patch = getPatchPreamble(masterDelta, reverse)
    preamblePartCount = len(patch)

    newHunkStartOffset = 0

    firstLinePos = lines[spanStart].hunkPos
    lastLinePos = lines[spanEnd].hunkPos

    for hunkID in range(firstLinePos.hunkID, lastLinePos.hunkID + 1):
        assert hunkID >= 0

        hunkStart, hunkLast = LineData.getHunkExtents(lines, hunkID)
        hunkHeader = lines[hunkStart]
        assert hunkHeader.hunkPos.isHunkHeaderLine()
        hunkStart += 1  # Skip header line
        hunkStop = hunkLast + 1

        # ---------------------------------------------------------------------
        # Compute selection bounds within the hunk

        if hunkID == firstLinePos.hunkID and not firstLinePos.isHunkHeaderLine():
            # Start of selection falls inside this hunk: Adjust slice start
            sliceStart = hunkStart + firstLinePos.hunkLineNum
        else:
            # Start of selection was before this hunk: Take entire hunk
            sliceStart = hunkStart

        if hunkID == lastLinePos.hunkID:
            if lastLinePos.isHunkHeaderLine():
                # End of selection falls inside hunk header: ignore this hunk
                break

            # End of selection falls inside current hunk: adjust slice end
            sliceStop = hunkStart + lastLinePos.hunkLineNum + 1
        else:
            # End of selection is beyond this hunk: Take entire hunk
            sliceStop = hunkStop

        # Selected line numbers in this hunk
        sliceRange = range(sliceStart, sliceStop)
        # Context line numbers above/below selection in this hunk
        headContextRange = range(hunkStart, sliceStart)
        tailContextRange = range(sliceStop, hunkStop)

        # Compute line count delta in this hunk
        lineCountDelta = sum(lines[ln].originDelta for ln in sliceRange)
        if reverse:
            lineCountDelta = -lineCountDelta

        # Skip this hunk if all selected lines are context
        if lineCountDelta == 0 and all(lines[ln].originDelta == 0 for ln in sliceRange):
            continue

        # ---------------------------------------------------------------------
        # Adapt hunk header

        # Parse hunk info
        hunkOldStart, hunkOldLines, hunkNewStart, hunkNewLines, hunkComment = hunkHeader.parseHunkHeader()

        # Get coordinates of old hunk
        if reverse:  # flip old<=>new if reversing
            oldStart, oldLines = hunkNewStart, hunkNewLines
        else:
            oldStart, oldLines = hunkOldStart, hunkOldLines

        # Compute coordinates of new hunk
        newStart = oldStart + newHunkStartOffset
        newLines = oldLines + lineCountDelta

        # Assemble doctored hunk header
        assert hunkComment.endswith("\n")
        patch.append(f"@@ -{oldStart},{oldLines} +{newStart},{newLines} @@")
        patch.append(hunkComment)

        # Account for line count delta in next new hunk's start offset
        newHunkStartOffset += lineCountDelta

        # ---------------------------------------------------------------------
        # Write hunk contents

        # Write non-selected lines at beginning of hunk as context
        writeContext(patch, reverse, (lines[ln] for ln in headContextRange))

        # We'll reorder all non-context lines so that "-" lines always appear above "+" lines.
        # This buffer will hold "+" lines while we're processing a clump of non-context lines.
        # This is to work around a libgit2 bug where it fails to parse "+" lines without LF
        # that appear above "-" lines. (Vanilla git doesn't have this issue.)
        # libgit2 fails to parse this:          But this parses fine:
        #   +hello                                -hallo
        #   \ No newline at end of file           +hello
        #   -hallo                                \ No newline at end of file
        buffer: list[str] = []

        # Write selected lines within the hunk
        for ln in sliceRange:
            line = lines[ln]
            origin = line.reverseOrigin if reverse else line.origin

            if origin == " " and buffer:
                # A context line breaks up the clump of non-context lines - flush buffer
                patch.extend(buffer)
                buffer.clear()

            if origin == "+":
                # Save "+" line for the end of the clump
                writeTo = buffer
            else:
                # Write "-" and " " lines to final patch now
                writeTo = patch
            writeTo.append(origin)
            writeTo.append(line.text)
            writeTo.append(line.hiddenSuffix)

        # Flush any remaining buffered "+" lines
        patch.extend(buffer)

        # End of selected lines.
        # All remaining lines in the hunk are context from now on.
        # Write non-selected lines at end of hunk as context
        writeContext(patch, reverse, (lines[ln] for ln in tailContextRange))

    # Bail if the patch comes out empty
    if len(patch) <= preamblePartCount:
        return ""

    return "".join(patch)



# -----------------------------------------------------------------------------
# Copyright (C) 2024 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.qt import *


class UrlToolTip(QTimer):
    def __init__(self, parent):
        super().__init__(parent)
        self._pendingText = ""
        self.setInterval(QApplication.style().styleHint(QStyle.StyleHint.SH_ToolTip_WakeUpDelay))
        self.setSingleShot(True)
        self.timeout.connect(lambda: QToolTip.showText(QCursor.pos(), self._pendingText))

    def linkHovered(self, url: str):
        self._pendingText = url
        if QToolTip.isVisible():
            self.stop()
            self.timeout.emit()
        else:
            self.start()

    def install(self):
        parentWidget = self.parent()
        assert isinstance(parentWidget, QWidget)
        for label in parentWidget.findChildren(QLabel):
            if label.openExternalLinks():
                label.linkHovered.connect(self.linkHovered)



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette import colors
from gitfourchette.localization import _
from gitfourchette.qt import *
from gitfourchette.toolbox import *

_SOB = "Signed-off-by"


class SignoffButton(QToolButton):
    def __init__(self, parent):
        super().__init__(parent)

        self.willSignoff = False

        self.setPopupMode(QToolButton.ToolButtonPopupMode.InstantPopup)

        toggleAction = QAction(self)
        toggleAction.setText(_("Add {0} to commit message", tquo(_SOB)))
        toggleAction.setCheckable(True)
        toggleAction.toggled.connect(lambda t: self.updateConfig(t))

        self.addActions([toggleAction])
        self.updateConfig(self.willSignoff)

    def updateConfig(self, willSignoff: bool):
        self.willSignoff = willSignoff

        lines = []

        if willSignoff:
            lines.append(_("A {0} trailer by the committer will be added "
                           "to the end of the commit message.", tquo(_SOB)))
        else:
            lines.append(_("{0} will <b>not</b> be added to the commit message.", tquo(_SOB)))

        tip = "<html style='white-space: pre-wrap;'>" + paragraphs(*lines)

        if willSignoff:
            icon = stockIcon("git-signoff", f"gray={colors.olive.name()}")
        else:
            icon = stockIcon("git-signoff-disabled")

        self.setIcon(icon)
        self.setToolTip(tip)

    def explicitSign(self):
        return self.willSignoff

    def explicitNoSign(self):
        return not self.willSignoff



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from .gitconflict import GitConflict
from .gitconflict import GitConflictSides
from .gitdelta import GitDelta, GitStatus
from .gitdeltafile import GitDeltaSource, GitDeltaFile
from .gitdriver import GitDriver
from .gitdriver import VanillaFetchStatusFlag
from .gitdriver import argsIf
from .lfspointer import LfsPointerState, LfsObjectCacheMissingError



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import enum
import os
from os import PathLike
from pathlib import Path

HOME = os.path.abspath(os.path.expanduser('~'))


class PathDisplayStyle(enum.IntEnum):
    FullPaths = 1
    AbbreviateDirs = 2
    FileNameOnly = 3
    FileNameFirst = 4


def compactPath(path: str | PathLike) -> str:
    # Normalize path first, which also turns forward slashes to backslashes on Windows.
    path = os.path.abspath(path)
    if path.startswith(HOME):
        path = "~" + path[len(HOME):]
    return path


def disambiguateTabTitlesByPath(workdirs: list[str]) -> list[str]:
    """Return unique tab titles for repos that share the same base name."""
    partsList = [Path(compactPath(workdir)).parts for workdir in workdirs]

    maxDepth = max(len(parts) for parts in partsList)
    uniqueDepth = maxDepth
    for depth in range(1, maxDepth + 1):
        tails = [parts[-depth:] for parts in partsList]
        if len(set(tails)) == len(tails):
            uniqueDepth = depth
            break

    labels = []
    for parts in partsList:
        tail = parts[-uniqueDepth:]
        if len(tail) == 1:
            labels.append(tail[0])
        elif len(tail) == 2:
            labels.append(str(Path(*tail)))
        else:
            labels.append(str(Path(tail[0], "…", tail[-1])))
    return labels


def abbreviatePath(
        path: str,
        style: PathDisplayStyle = PathDisplayStyle.FullPaths,
        allowNul: bool = False
) -> str:
    if style == PathDisplayStyle.AbbreviateDirs:
        splitLong = path.split('/')
        for i in range(len(splitLong) - 1):
            if splitLong[i][0] == '.':
                splitLong[i] = splitLong[i][:2]
            else:
                splitLong[i] = splitLong[i][0]
        return '/'.join(splitLong)

    elif style == PathDisplayStyle.FileNameFirst:
        try:
            directory, file = path.rsplit('/', 1)
            separator = '\0 ' if allowNul else '  '
            return separator.join((file, directory))
        except ValueError:
            return path

    elif style == PathDisplayStyle.FileNameOnly:
        return path.rsplit('/', 1)[-1]

    else:
        return path



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

import weakref

from gitfourchette import settings
from gitfourchette.application import GFApplication
from gitfourchette.localization import *
from gitfourchette.qt import *
from gitfourchette.toolbox import stockIcon
from gitfourchette.toolbox.qtutils import CallbackAccumulator, reevaluateStyleSheet


class QTabBar2(QTabBar):
    tabMiddleClicked = Signal(int)
    tabDoubleClicked = Signal(int)
    visibilityChanged = Signal(bool)
    wheelDelta = Signal(QPoint)
    layoutChanged = Signal()
    suggestScrollToCurrentTab = Signal()

    middleClickedIndex: int
    doubleClickedIndex: int
    shadowText: list[str]

    def __init__(self, parent: QWidget):
        super().__init__(parent)
        self.middleClickedIndex = -1
        self.doubleClickedIndex = -1
        self.setObjectName("QTabBar2")
        self.shadowText = []

        self.tabMoved.connect(self._onTabMoved)

    # -------------------------------------------------------------------------
    # Layout

    def tabSizeHint(self, index: int) -> QSize:
        # Keep individual tabs from getting too wide when several tabs compete
        # for real estate. Qt may ignore this hint when the tab bar is wide
        # enough to fit all tabs.

        try:
            # Avoid tabText() here - it may return pre-elided text!
            text = self.shadowText[index]
        except IndexError:
            # May occur until tabInserted is called
           text = self.tabText(index)

        hint = super().tabSizeHint(index)

        w = 32 + self.fontMetrics().horizontalAdvance(text)

        if self.tabsClosable():
            closeWidth = self.style().pixelMetric(QStyle.PixelMetric.PM_TabCloseIndicatorWidth)
            w += 2 * closeWidth

        w = min(w, 250)

        hint.setWidth(w)
        return hint

    def tabLayoutChange(self):
        self.layoutChanged.emit()

    # -------------------------------------------------------------------------
    # Mouse

    def mousePressEvent(self, event: QMouseEvent):
        if event.button() == Qt.MouseButton.MiddleButton:
            self.middleClickedIndex = self.tabAt(event.position().toPoint())
        else:
            self.middleClickedIndex = -1
        super().mousePressEvent(event)

    def mouseDoubleClickEvent(self, event: QMouseEvent):
        if event.button() == Qt.MouseButton.LeftButton:
            self.doubleClickedIndex = self.tabAt(event.position().toPoint())
        else:
            self.doubleClickedIndex = -1
        super().mouseDoubleClickEvent(event)

    def mouseMoveEvent(self, event: QMouseEvent):
        # Block double-click signal if mouse moved before releasing button
        self.doubleClickedIndex = -1
        super().mouseMoveEvent(event)

    def wheelEvent(self, event: QWheelEvent):
        self.wheelDelta.emit(event.angleDelta())

        # DO NOT forward mouse wheel events to superclass
        # (avoid default Qt behavior that switches tabs via scroll wheel)
        event.accept()

    def mouseReleaseEvent(self, event: QMouseEvent):
        runStandardHandler = True

        if event.buttons() != Qt.MouseButton.NoButton:
            # Signals will only fire if clicking a single button at a time
            pass

        elif event.button() == Qt.MouseButton.MiddleButton:
            i = self.tabAt(event.position().toPoint())
            if i >= 0 and i == self.middleClickedIndex:
                # Bypass standard mouseReleaseEvent to avoid hardcoded behavior
                # on PySide6+GNOME where middle-clicking a tab emits
                # tabCloseRequested, if tabsClosable is enabled.
                runStandardHandler = False
                self.tabMiddleClicked.emit(i)

        elif event.button() == Qt.MouseButton.LeftButton:
            i = self.tabAt(event.position().toPoint())
            if i >= 0 and i == self.doubleClickedIndex:
                runStandardHandler = False
                self.tabDoubleClicked.emit(i)

        self.middleClickedIndex = -1
        self.doubleClickedIndex = -1

        if runStandardHandler:
            super().mouseReleaseEvent(event)

    # -------------------------------------------------------------------------
    # Misc. events

    def focusInEvent(self, event: QFocusEvent):
        """Suggest scrolling to current tab when we receive keyboard focus"""
        super().focusInEvent(event)
        self.suggestScrollToCurrentTab.emit()

    def setVisible(self, visible: bool):
        """Forward setVisible to parent scroll area (when we get hidden due to only 1 tab remaining)"""
        super().setVisible(visible)
        self.visibilityChanged.emit(visible)

    # -------------------------------------------------------------------------
    # Shadow text bookkeeping - Needed because calling tabText() inside
    # tabSizeHint() returns elided text!

    def _onTabMoved(self, i: int, j: int):
        del self.shadowText[i]
        self.shadowText.insert(j, self.tabText(j))

    def tabInserted(self, index):
        self.shadowText.insert(index, self.tabText(index))
        super().tabInserted(index)

    def tabRemoved(self, index):
        del self.shadowText[index]
        super().tabRemoved(index)

    def setTabText(self, index: int, text: str):
        currentText = self.shadowText[index]

        # Overriding tab text can be expensive, so avoid if there's no change
        if currentText != text:
            self.shadowText[index] = text
            super().setTabText(index, text)


class QTabWidget2OverflowGradient(QWidget):
    def paintEvent(self, event: QPaintEvent):
        W = 16  # gradient width
        P = 4  # opaque padding inside the gradient

        scrollArea = self.parentWidget()
        assert isinstance(scrollArea, QAbstractScrollArea)
        scrollBar = scrollArea.horizontalScrollBar()

        saw = scrollArea.width()
        sah = scrollArea.height()

        opaque = self.palette().color(QPalette.ColorRole.Window)
        transp = QColor(opaque)
        transp.setAlpha(0)

        painter = QPainter(self)

        if scrollBar.value() != 0:
            gradient = QLinearGradient(P, 0, W, 0)
            gradient.setColorAt(0, opaque)
            gradient.setColorAt(1, transp)
            painter.fillRect(QRect(0, 0, W, sah-1), gradient)

        if scrollBar.value() < scrollBar.maximum()-1:
            gradient = QLinearGradient(saw-W, 0, saw-P, 0)
            gradient.setColorAt(0, transp)
            gradient.setColorAt(1, opaque)
            painter.fillRect(QRect(saw-W, 0, W, sah-1), gradient)


class QTabWidget2(QWidget):
    currentChanged: Signal = Signal(int)
    tabCloseRequested: Signal = Signal(int)
    tabMiddleClicked: Signal = Signal(int)
    tabDoubleClicked: Signal = Signal(int)
    tabContextMenuRequested: Signal = Signal(QPoint, int)

    currentWidgetChanged = Signal()
    """Emitted when the displayed widget is actually changed (in contrast to
    currentChanged, which is emitted even when dragging the foreground tab)."""

    UrgentPropertyName = "QTabBar2_UrgentFlag"

    def __init__(self, parent):
        super().__init__(parent)
        self.setObjectName("QTabWidget2")

        self.stacked = QStackedWidget(self)
        self.stacked.setObjectName("QTW2StackedWidget")

        self.shadowCurrentWidgetRef: weakref.ref[QWidget] = weakref.ref(self)
        """ Keep a weakref instead of a real reference to not impede GC of a
        dead widget. (A weakref on self stands in for "none".) """

        self.tabScrollArea = QScrollArea(self)
        self.tabScrollArea.setWidgetResizable(True)
        self.tabScrollArea.setFrameStyle(QFrame.Shape.NoFrame)
        self.tabScrollArea.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.tabScrollArea.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.tabScrollArea.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        self.tabScrollArea.setFocusPolicy(Qt.FocusPolicy.NoFocus)

        self.tabs = QTabBar2(self.tabScrollArea)
        self.tabs.tabMoved.connect(self.onTabMoved)
        self.tabs.currentChanged.connect(self.onCurrentChanged)
        self.tabs.tabCloseRequested.connect(self.tabCloseRequested)
        self.tabs.tabMiddleClicked.connect(self.tabMiddleClicked)
        self.tabs.tabDoubleClicked.connect(self.tabDoubleClicked)
        self.tabs.suggestScrollToCurrentTab.connect(self.ensureCurrentTabVisible)
        self.tabs.setMovable(True)
        self.tabs.setDocumentMode(True)  # dramatically improves the tabs' appearance on macOS
        self.tabs.setUsesScrollButtons(False)  # can't have those with scroll area
        self.tabs.setElideMode(Qt.TextElideMode.ElideRight)

        self.overflowButton = QToolButton(self)
        self.overflowButton.setArrowType(Qt.ArrowType.DownArrow)
        self.overflowButton.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.overflowButton.setToolTip(_("List all tabs"))
        self.overflowButton.clicked.connect(self.onOverflowButtonClicked)
        self.overflowButton.setCheckable(True)
        self.overflowButton.setAutoRaise(True)
        self.overflowButton.hide()  # hiding now prevents jitter on boot because maximum height is adjuster later

        self.tabScrollArea.setWidget(self.tabs)
        self.tabs.visibilityChanged.connect(self.tabScrollArea.setVisible)
        self.tabs.wheelDelta.connect(self.scrollTabs)
        self.tabs.layoutChanged.connect(self.updateOverflowDropdown)

        self.overflowGradient = QTabWidget2OverflowGradient(self.tabScrollArea)
        self.overflowGradient.setObjectName("QTW2OverflowGradient")
        self.overflowGradient.setAttribute(Qt.WidgetAttribute.WA_TransparentForMouseEvents)

        topWidget = QWidget(self)
        self.topWidget = topWidget
        topLayout = QHBoxLayout(topWidget)
        topLayout.setSpacing(2)
        topLayout.setContentsMargins(0, 0, 0, 0)
        topLayout.addWidget(self.tabScrollArea)
        topLayout.addWidget(self.overflowButton)

        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(topWidget)
        layout.addWidget(self.stacked)

        self.tabs.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.tabs.customContextMenuRequested.connect(self.onCustomContextMenuRequested)

        self.overflowMenu = QMenu(self)
        self.overflowMenu.setObjectName("QTW2OverflowMenu")
        self.overflowMenu.setToolTipsVisible(True)
        self.overflowMenu.aboutToHide.connect(lambda: self.overflowButton.setChecked(False))

        GFApplication.instance().prefsChanged.connect(self.refreshPrefs)
        self.refreshPrefs()

    def __len__(self):
        return self.tabs.count()

    def refreshPrefs(self):
        mustReevaluateStyleSheet = settings.prefs.expandingTabs != self.tabs.expanding()

        self.tabs.setExpanding(settings.prefs.expandingTabs)
        self.tabs.setAutoHide(settings.prefs.autoHideTabs)
        self.tabs.setTabsClosable(settings.prefs.tabCloseButton)
        self.tabs.update()
        self.syncBarSize()
        self.onResize()
        self.updateOverflowDropdown()

        if mustReevaluateStyleSheet:
            reevaluateStyleSheet(self)

    def onCustomContextMenuRequested(self, localPoint: QPoint):
        globalPoint = self.tabs.mapToGlobal(localPoint)
        index = self.tabs.tabAt(localPoint)
        self.tabContextMenuRequested.emit(globalPoint, index)

    def onTabMoved(self, fromIndex: int, toIndex: int):
        # Keep QStackedWidget in sync with QTabBar
        w = self.stacked.widget(fromIndex)
        self.stacked.removeWidget(w)
        self.stacked.insertWidget(toIndex, w)

    def onCurrentChanged(self, i: int):
        # Keep QStackedWidget in sync with QTabBar
        self.stacked.setCurrentIndex(i)
        currentWidget = self.currentWidget()

        # Make sure the tab is visible within the scrollable area
        self.ensureCurrentTabVisible()

        # Remove urgent flag if any
        if currentWidget is not None and currentWidget.property(QTabWidget2.UrgentPropertyName):
            currentWidget.setProperty(QTabWidget2.UrgentPropertyName, None)
            self.tabs.setTabIcon(i, QIcon())  # clear icon

        # See if we should emit the currentWidgetChanged signal
        currentWidgetRef = weakref.ref(currentWidget or self)  # self stands in for None
        if currentWidgetRef != self.shadowCurrentWidgetRef:
            self.shadowCurrentWidgetRef = currentWidgetRef
            self.currentWidgetChanged.emit()

        # Forward signal
        self.currentChanged.emit(i)

    def addTab(self, w: QWidget, name: str) -> int:
        return self.insertTab(self.count(), w, name)

    def insertTab(self, index: int, w: QWidget, name: str) -> int:
        i1 = self.stacked.insertWidget(index, w)
        i2 = self.tabs.insertTab(index, name)
        self.syncBarSize()
        assert i1 == i2
        return i1

    def setCurrentIndex(self, i: int):
        self.tabs.setCurrentIndex(i)
        self.stacked.setCurrentIndex(i)
        self.tabs.update()

    def setCurrentWidget(self, widget: QWidget):
        i = self.indexOf(widget)
        assert i >= 0
        self.setCurrentIndex(i)

    def indexOf(self, widget: QWidget):
        return self.stacked.indexOf(widget)

    def widget(self, i: int):
        return self.stacked.widget(i)

    def currentIndex(self) -> int:
        return self.stacked.currentIndex()

    def setTabText(self, i: int, text: str):
        self.tabs.setTabText(i, text)

    def setTabTooltip(self, i: int, toolTip: str):
        self.tabs.setTabToolTip(i, toolTip)

    def currentWidget(self) -> QWidget:
        return self.stacked.currentWidget()

    def count(self) -> int:
        return self.stacked.count()

    def removeTab(self, i: int):
        assert isinstance(i, int)
        widget = self.stacked.widget(i)
        # remove widget from stacked view _before_ removing the tab,
        # because removing the tab may send a tab change event
        self.stacked.removeWidget(widget)
        self.tabs.removeTab(i)
        self.syncBarSize()

    def swapWidget(self, i: int, newWidget: QWidget):
        currentIndex = self.currentIndex()
        oldWidget = self.stacked.widget(i)
        self.stacked.removeWidget(oldWidget)
        assert oldWidget.parentWidget() is self.stacked, "QStackedWidget isn't supposed to deparent its pages"
        self.stacked.insertWidget(i, newWidget)
        self.stacked.setCurrentIndex(currentIndex)  # Keep it stable
        if currentIndex == i:
            self.shadowCurrentWidgetRef = weakref.ref(newWidget)
            self.currentWidgetChanged.emit()
        return oldWidget

    def widgets(self):
        for i in range(self.stacked.count()):
            yield self.stacked.widget(i)

    def syncBarSize(self):
        h = self.tabs.sizeHint().height()
        if h != 0:
            self.tabScrollArea.setFixedHeight(self.tabs.sizeHint().height())
            self.overflowGradient.resize(self.tabScrollArea.size())

    # I don't like deferring ensureCurrentTabVisible to the next event loop,
    # but the new tabbar width doesn't seem to be refreshed immediately in onCurrentChanged.
    @CallbackAccumulator.deferredMethod()
    def ensureCurrentTabVisible(self):
        i = self.currentIndex()
        if i < 0:
            return
        rect = self.tabs.tabRect(i)

        vr = self.tabScrollArea.viewport().contentsRect()
        vr.translate(self.tabScrollArea.horizontalScrollBar().value(), self.tabScrollArea.verticalScrollBar().value())

        if rect.left() < vr.left():
            p = QPoint(rect.left(), rect.center().y())
        else:
            p = QPoint(rect.right(), rect.center().y())
        self.tabScrollArea.ensureVisible(p.x(), p.y())

    def scrollTabs(self, delta: QPoint):
        x = delta.x()
        y = delta.y()
        xBias = abs(x) - abs(y)

        # Reduce jitter when it's unclear which axis the user intends to scroll on
        # (for trackpads)
        if abs(xBias) < 1:
            return

        deltaValue = x if xBias > 0 else y
        scrollBar = self.tabScrollArea.horizontalScrollBar()
        scrollBar.setValue(scrollBar.value() - deltaValue)

    @CallbackAccumulator.deferredMethod()  # don't update overflow button too often
    def updateOverflowDropdown(self):
        if self.tabs.count() <= 1:  # never overflow if there's just one tab
            isOverflowing = False
        else:
            isOverflowing = self.topWidget.width() < self.tabs.width()

        self.overflowGradient.setVisible(isOverflowing)
        self.overflowButton.setVisible(isOverflowing)

        if isOverflowing:
            self.overflowButton.setMaximumHeight(self.tabs.height())

    def onOverflowButtonClicked(self):
        self.overflowMenu.clear()
        for i in range(self.count()):
            action = QAction(self.overflowMenu)
            action.setText(self.tabs.tabText(i))
            action.setToolTip(self.tabs.tabToolTip(i))
            action.triggered.connect(lambda _dummy, j=i: self.setCurrentIndex(j))
            self.overflowMenu.addAction(action)

        pos = self.mapToGlobal(self.overflowButton.pos() + self.overflowButton.rect().bottomLeft())
        self.overflowMenu.popup(pos)

    def resizeEvent(self, event: QResizeEvent):
        super().resizeEvent(event)
        self.onResize()

    def onResize(self):
        self.overflowGradient.resize(self.tabScrollArea.size())
        self.ensureCurrentTabVisible()

    def requestAttention(self, i: int):
        if i == self.currentIndex():
            return
        widget = self.widget(i)
        if widget is None:
            return
        if widget.property(QTabWidget2.UrgentPropertyName) == "true":
            return
        widget.setProperty(QTabWidget2.UrgentPropertyName, "true")
        self.tabs.setTabIcon(i, stockIcon("urgent-tab"))



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from .colorscheme import ColorScheme, PygmentsPresets
from .lexercache import LexerCache
from .lexjob import LexJob
from .lexjobcache import LexJobCache



# -----------------------------------------------------------------------------
# Copyright (C) 2026 Iliyas Jorio.
# This file is part of GitFourchette, distributed under the GNU GPL v3.
# For full terms, see the included LICENSE file.
# -----------------------------------------------------------------------------

from gitfourchette.qt import *


class QSignalConnectContext:
    def __init__(self, signal: SignalInstance, slot):
        self.signal = signal
        self.slot = slot

    def __enter__(self):
        self.connection = self.signal.connect(self.slot)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.signal.disconnect(self.connection)
